from typing import *
import pathlib
import abc

from base_aux.classes import AnnotAux
from . import Exx__FileNotExists, TYPE__PATH


# =====================================================================================================================
# TODO: add iter???


# =====================================================================================================================
class PrivateBase(AnnotAux, abc.ABC):
    """Base class to get values from sources.

    :ivar SECTION: first level name in source, for ini - root section, for json - rootKey, for env - not used
    :ivar DIRPATH: file destination
    :ivar FILENAME: file name

    USAGE
    -----
    if you dont need RAISE when no value get for exact annotated name - just define None!
    """
    _RAISE: Optional[bool] = True

    SECTION: Optional[str] = None

    DIRPATH: Optional[TYPE__PATH] = pathlib.Path.home()
    FILENAME: Optional[str] = None

    _text: Optional[str] = ""     # TODO: need tests!!!
    dict: Dict[str, Any] = None

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(
            self,
            _section: Optional[str] = None,

            _dirpath: TYPE__PATH = None,
            _filename: str = None,
            _filepath: TYPE__PATH = None,

            _text: Optional[str] = None,                # instead of file
            _dict: Optional[Dict[str, Any]] = None,     # instead of file

            _raise: Optional[bool] = None
    ):
        super().__init__()
        self.SECTION = _section or self.SECTION

        if _dict is not None:
            self.SECTION = None
            self.DIRPATH = None
            self.FILENAME = None
            self.apply_dict(_dict)
        elif _text is not None:
            self.DIRPATH = None
            self.FILENAME = None
            self._text = _text
        else:
            if not self._filepath_apply_new(
                _dirpath=_dirpath,
                _filename=_filename,
                _filepath=_filepath
            ):
                return

        if _dict is None:
            self.load_dict()

        if _raise is not None:
            self._RAISE = _raise

        if self._RAISE:
            self.annot__raise_if_not_defined()

    def __str__(self):
        """return pretty string
        """
        result = f"{self.filepath=}"
        data = self.get_dict()
        if data:
            for key, value in data.items():
                result += f"\n{key}={value}"
        elif self.filepath and self.filepath.exists():
            result += f"\n{self._text}"
        else:
            result += f"\ndata=None"
        return result

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def file_check_exists(cls) -> bool:
        """
        SPECIALLY CREATED FOR
        ---------------------
        for pytest when i use Privates
        i need to skip tests when i have no Privates! like in githubActions or on workPC

        by now just use
            @pytest.mark.skipif(Lambda().check_raise(), reason="no file")
        """
        pass
        # FIXME: finish when pigs fly!

    def _filepath_apply_new(
            self,
            _dirpath: TYPE__PATH = None,
            _filename: str = None,
            _filepath: TYPE__PATH = None
    ) -> bool:
        """apply new file destination
        """
        if not _filepath:
            self.DIRPATH = pathlib.Path(_dirpath or self.DIRPATH)
            self.FILENAME = _filename or self.FILENAME
        else:
            self.DIRPATH = pathlib.Path(_filepath).parent
            self.FILENAME = pathlib.Path(_filepath).name

        if self.filepath and not self.filepath.exists():
            if self._RAISE:
                msg = f'[CRITICAL]no[{self.filepath=}]'
                raise Exx__FileNotExists(msg)
            else:
                return False

        if self.filepath:
            self._text = self.filepath.read_text()

        return True

    @property
    def filepath(self) -> Optional[pathlib.Path]:
        """compose final pathlib instance for file
        """
        try:
            if self.FILENAME:
                return pathlib.Path(self.DIRPATH, self.FILENAME)
        except:
            pass

    # -----------------------------------------------------------------------------------------------------------------
    def load_dict(self) -> None:
        """load values from source into instance attributes.
        """
        section_dict = self.get_dict()
        self.apply_dict(section_dict)

    def apply_dict(self, attrs: Optional[Dict[str, Any]] = None, update: Optional[bool] = None) -> None | NoReturn:
        """Apply passes dict into instance and check consistence.
        """
        # clear -----------------------
        if self.dict and not update:
            for key in self.dict:
                delattr(self, key)

        # work -----------------------
        if attrs is not None:
            if update:
                self.dict.update(attrs)
            else:
                self.dict = dict(attrs)

        if self.dict is None:
            return

        for key, value in self.dict.items():
            setattr(self, key, value)

    def update_dict(self, attrs: Optional[Dict[str, Any]]) -> None | NoReturn:
        """Apply passes dict into instance and check consistence.
        """
        self.apply_dict(attrs, True)

    def preupdate_dict(self, attrs: Dict[str, Any]) -> None | NoReturn:
        """Apply passes dict into instance and check consistence.
        """
        # prepare  -----------------------
        new_dict = {}
        for key, value in attrs.items():
            if key not in self.dict:
                new_dict.update({key: value})

        # work -----------------------
        self.update_dict(new_dict)

    # -----------------------------------------------------------------------------------------------------------------
    @abc.abstractmethod
    def get_dict(self) -> Optional[Dict[str, Any]]:
        """Obtain existed values from source in dict structure.

        return
            NONE - if no section! dont raise inside!
            {} - if no names!
        """
        pass


# =====================================================================================================================



from typing import *
import re

from . import ValidAux


# =====================================================================================================================
class ValidRegExp:
    """
    GOAL
    ----
    use ready object to validate by comparing with others

    SPECIALLY CREATED FOR
    ---------------------
    buses.serial_client.SerialClient.write_read__last_validate_regexp
    replace write_read__last_validate_regexp by using only write_read__last_validate with this object as expected.

    BEST USAGE
    ----------
    OBJECT cmp
        assert 1 == ValidRegExp(r"\\d?")
        assert 1 == ValidRegExp([r"\\d?", r"\\s*\\d?"])
    METHOD check
        assert ValidRegExp(r"\\d?").run(1)

    """
    REGEXPS: Iterable[str] = None
    IGNORECASE: bool = True

    def __init__(self, regexps: str | Iterable[str] = None, ignorecase: bool = None):
        if ignorecase is not None:
            self.IGNORECASE = ignorecase

        if regexps:
            self.REGEXPS = regexps

        if isinstance(self.REGEXPS, str):
            self.REGEXPS = [self.REGEXPS, ]

    def __eq__(self, other: Any) -> bool:
        if not self.REGEXPS:
            return True

        for pattern in self.REGEXPS:
            if re.fullmatch(pattern=str(pattern), string=str(other), flags=re.RegexFlag.IGNORECASE if self.IGNORECASE else 0):
                return True
        return False

    def run(self, source: Any) -> bool:
        return ValidAux.compare_doublesided__bool(source, self)


# =====================================================================================================================



from typing import *
import time

from .valid_0_aux import ValidAux

from base_aux.objects import TypeChecker
from base_aux.funcs import *
from ..funcs.static import TYPE__VALID_VALIDATOR


# =====================================================================================================================
class Valid(ValidAux):
    """
    GOAL
    ----
    1. get value from somewhere and clearly validate it.
    2. ability to log result in appropriate way by using template-pattern for msg log

    CREATED SPECIALLY FOR
    ---------------------
    testplans sequences or same staff

    CONSTRAINTS
    -----------
    1. Exceptions
    any exception is a special state!
    so dont try to compare Exception with any real NoExx value or validate it in validator_link!
    if expect exception just place it directly in Validator!

    2. Callables
    only funcs/methods will be called
    if Class - no call! no init!

    3. VALIDATE
    if final value - it will be compared directly by compare_doublesided.
    So be careful about True (as default) and apply LAMBDA_BOOL(VALUE) if you need bool(val) comparing!

    BEST USAGE
    ----------

    WHY NOT: 1?
    -----------

    WHY NOT: 2?
    -----------


    ATTRS
    =====
    :ivar REVERSE_LINK:
    created specially for applying DRY in testplans - then PSU will not PowerOk on different HvIn
    """

    NAME: str = ""      # TODO: realise access to Valid from Chains!
    COMMENT: str = ""

    SKIP_LINK: TYPE__VALID_SOURCE_BOOL = None
    VALUE_LINK: TYPE__VALID_SOURCE
    VALIDATE_LINK: TYPE__VALID_VALIDATOR = True
    VALIDATE_RETRY: int = 0
    REVERSE_LINK: TYPE__VALID_SOURCE_BOOL = None    # in case of REVERSE - REAL RESULT IS validate_last_bool!!! idea is validate_last have direct validationResult but reversing goes into validate_last_bool

    ARGS__VALUE: TYPE__VALID_ARGS = ()
    ARGS__VALIDATE: TYPE__VALID_ARGS = ()
    KWARGS__VALUE: TYPE__VALID_KWARGS = None
    KWARGS__VALIDATE: TYPE__VALID_KWARGS = None

    # RESULT ACTUAL ------------------------------
    timestamp_last: float | None = None
    skip_last: TYPE__VALID_RESULT_BOOL = False
    finished: bool | None = None
    value_last: TYPE__VALID_RESULT = None               # direct result value for calculating func value_link
    validate_last: TYPE__VALID_RESULT_BOOL__EXX = True   # direct result value for calculating func validate_link === decide using only bool???
    reverse_last: TYPE__VALID_RESULT_BOOL = None
    validate_last_bool: TYPE__VALID_RESULT_BOOL         # represented value for validation
    log_lines: list[str] = None

    # CHAINS -------------------------------------
    CHAIN__CUM: bool = True
    CHAIN__FAIL_STOP: bool = True

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(
            self,
            value_link: TYPE__VALID_SOURCE = ValueNotExist,
            validate_link: Optional[TYPE__VALID_VALIDATOR] = None,
            validate_retry: Optional[int] = None,
            skip_link: TYPE__VALID_SOURCE_BOOL= None,
            reverse_link: TYPE__VALID_SOURCE_BOOL = None,

            args__value: TYPE__VALID_ARGS = (),
            args__validate: TYPE__VALID_ARGS = (),

            kwargs__value: TYPE__VALID_KWARGS = None,
            kwargs__validate: TYPE__VALID_KWARGS = None,

            name: Optional[str] = None,
            comment: Optional[str] = None,

            chain__cum: Optional[bool] = None,
            chain__fail_stop: Optional[bool] = None,
    ):
        """
        :param value_link: None - for cmp by eq/ne! other types - for direct usage
        :param validate_link:
        :param skip_link:
        :param args__value:
        :param args__validate:
        :param kwargs__value:
        :param kwargs__validate:
        :param name:
        :param comment:
        :param chain__cum:
        :param chain__fail_stop:
        """
        self.clear()

        self.VALUE_LINK = value_link

        if validate_link is not None:
            self.VALIDATE_LINK = validate_link
        if validate_retry is not None:
            self.VALIDATE_RETRY = validate_retry
        if reverse_link is not None:
            self.REVERSE_LINK = reverse_link
        if skip_link is not None:
            self.SKIP_LINK = skip_link

        # ARGS/KWARGS --------------------------------
        self.ARGS__VALUE = args__ensure_tuple(args__value)
        self.ARGS__VALIDATE = args__ensure_tuple(args__validate)

        self.KWARGS__VALUE = kwargs__value or {}
        self.KWARGS__VALIDATE = kwargs__validate or {}

        # INFO ---------------------------------------
        if name:
            self.NAME = name
        if comment:
            self.COMMENT = comment

        # CHAINS -------------------------------------
        if chain__cum is not None:
            self.CHAIN__CUM = chain__cum
        if chain__fail_stop is not None:
            self.CHAIN__FAIL_STOP = chain__fail_stop

    def clear(self):
        self.timestamp_last = None
        self.skip_last = False
        self.finished = None
        self.value_last = None
        self.validate_last = True
        self.log_lines = []

    @property
    def validate_last_bool(self) -> bool:
        """
        this is link name for bool(result)
        """
        return bool(self)

    def get_finished_result_or_none(self) -> None | bool:
        """
        GOAL
        ----
        attempt to make ability to get clear result by one check (for Gui)

        :return: if not finished - None
            if finished - validate_last_bool
        """
        if self.finished is None:
            return None
        else:
            return self.validate_last_bool

    def run__if_not_finished(self) -> bool:
        if not self.finished:
            return self.run()
        else:
            return bool(self)

    def run(self, value_link: Any = ValueNotExist) -> bool:
        """
        CONSTRAINTS
        -----------
        careful about 1 comparing (assert 0 == False, assert 1 == True, assert 2 != True)

        :param value_link: BE CAREFUL created specially for value_link=ValueNotExist on init
        """
        if value_link == ValueNotExist:
            value_link = self.VALUE_LINK

        self.clear()
        self.timestamp_last = time.time()

        # SKIP ---------------------
        self.skip_last = self.get_result_bool(self.SKIP_LINK)

        if not self.skip_last:
            retry_count = 0
            # WORK =======================
            self.finished = False

            while True:
                self.clear()
                self.timestamp_last = time.time()

                # VALUE ---------------------
                self.value_last = self.get_result_or_exx(value_link, args=self.ARGS__VALUE, kwargs=self.KWARGS__VALUE)

                # VALIDATE ------------------
                if isinstance(self.value_last, Exception) and not TypeChecker.check__exception(self.VALIDATE_LINK):
                    self.validate_last = False

                elif TypeChecker.check__exception(self.VALIDATE_LINK):
                    self.validate_last = TypeChecker.check__nested__by_cls_or_inst(self.value_last, self.VALIDATE_LINK)

                elif TypeChecker.check__callable_func_meth_inst(self.VALIDATE_LINK):
                    args_validate = (self.value_last, *self.ARGS__VALIDATE)
                    self.validate_last = self.get_result_or_exx(self.VALIDATE_LINK, args=args_validate, kwargs=self.KWARGS__VALIDATE)

                else:
                    self.validate_last = self.compare_doublesided_or_exx(self.value_last, self.VALIDATE_LINK)

                # FINISH retry
                if not self.VALIDATE_RETRY or retry_count == self.VALIDATE_RETRY or self.validate_last_bool:
                    break
                else:
                    retry_count += 1

            if self.REVERSE_LINK:
                self.reverse_last = self.get_result_bool(self.REVERSE_LINK)

            self.finished = True
            # ============================

        # FINISH final ---------------------
        return self.validate_last_bool

    # def validate(self, value_link: Any = ValueNotExist) -> bool:

    def __bool__(self) -> bool:
        if not self.finished:
            return False

        if self.reverse_last:
            return self.validate_last != True       # dont use validate_last_bool!!! recursion!
        else:
            return self.validate_last == True

    def __str__(self) -> str:
        # main ---------------
        # # TODO: apply name from source!!! if not passed???
        # STR_PATTERN: str = (
        #     "{0.__class__.__name__}(validate_last_bool={0.validate_last_bool},validate_last={0.validate_last},\n"
        #     "...VALUE_LINK={0.VALUE_LINK},ARGS__VALUE={0.ARGS__VALUE},KWARGS__VALUE={0.KWARGS__VALUE},value_last={0.value_last},\n"
        #     "...VALIDATE_LINK={0.VALIDATE_LINK},ARGS__VALIDATE={0.ARGS__VALIDATE},KWARGS__VALIDATE={0.KWARGS__VALIDATE},\n"
        #     "...skip_last={0.skip_last},NAME={0.NAME},finished={0.finished},timestamp_last={0.timestamp_last})"
        # )
        # result_str = STR_PATTERN.format(self)

        # -------------------------------------
        result_str = f"{self.__class__.__name__}(NAME={self.NAME},skip_last={self.skip_last},validate_last_bool={self.validate_last_bool}"
        if self.reverse_last:
            result_str += f",reverse_last={self.reverse_last}"
        result_str += f",\n"

        # value ----
        result_str += f"...VALUE_LINK={self.VALUE_LINK}"
        if self.ARGS__VALUE:
            result_str += f",ARGS__VALUE={self.ARGS__VALUE}"
        if self.KWARGS__VALUE:
            result_str += f",KWARGS__VALUE={self.KWARGS__VALUE}"
        result_str += f",value_last={self.value_last},\n"

        # validate ----
        result_str += f"...VALIDATE_LINK={self.VALIDATE_LINK}"
        if self.ARGS__VALIDATE:
            result_str += f",ARGS__VALIDATE={self.ARGS__VALIDATE}"
        if self.KWARGS__VALIDATE:
            result_str += f",KWARGS__VALIDATE={self.KWARGS__VALIDATE}"

        result_str += f",validate_last={self.validate_last}"
        if self.REVERSE_LINK:
            result_str += f"*REVERSE_LINK={self.REVERSE_LINK}"
        result_str += f",\n"

        # finish ----
        result_str += f",finished={self.finished},timestamp_last={self.timestamp_last},"

        # log ----------------
        for index, line in enumerate(self.log_lines):
            result_str += "\n" + f"{index}:".rjust(5, '_') + line

        return result_str

    def __repr__(self) -> str:
        return str(self)

    # -----------------------------------------------------------------------------------------------------------------
    def __eq__(self, other) -> bool:
        """
        GOAL
        ----
        use created object (with not defined value_link=None) as validator by EQ/NE inline methods!

        USAGE
        -----
        assert 1.0 >= 1
        assert float("1.0") >= 1
        assert "1.0" == Valid(validate_link=lambda x: float(x) >= 1)

        SPECIALLY CREATED FOR
        --------------------
        test uart devises by schema!

        assert "220.0V" == ValueUnit(...)
        assert "OFF" == ValueVariant("OFF", ["OFF", "ON"])
        assert "1.0" == Valid(validate_link=lambda x: float(x) >= 1)

        ValidTypeFloat = Valid(validate_link=lambda x: isinstance(x, float))
        assert "1.0" != ValidTypeFloat()
        assert 1.0 == ValidTypeFloat()
        """
        if self.VALUE_LINK == ValueNotExist:
            return self.run(other)
        else:
            # todo: maybe its not so good here/need ref? - seems OK!
            return self.run__if_not_finished() == other


# =====================================================================================================================



import pytest

from base_aux.monitors import *
from base_aux.alerts import *


# =====================================================================================================================
# TODO: create full auto test!


# =====================================================================================================================
@pytest.mark.skipif(AlertSmtp.AUTH.check_raise(), reason="no file")
class Test_UrlTag:
    def test__1(self):
        Monitor_DonorSvetofor()

    def test__2(self):
        Monitor_CbrKeyRate()

    def test__3(self):
        Monitor_ConquestS23_comments()

    def test__4(self):
        Monitor_Sportmaster_AdidasSupernova2M()

    def test__5(self):
        Monitor_AcraRaiting_GTLC()


# =====================================================================================================================
@pytest.mark.skipif(AlertSmtp.AUTH.check_raise(), reason="no file")
@pytest.mark.parametrize(argnames="pattern", argvalues=[None, r"\[ALERT\]test1"])
def test__imap(pattern):
    _subj_name = "test1"

    AlertSmtp(_subj_name=_subj_name)

    victim = MonitorImap
    victim.stop_flag = True
    victim.SUBJECT_REGEXP = pattern

    for i in range(3):
        victim_inst = victim()
        victim_inst.wait_cycle()
        if f"[ALERT]{_subj_name}" in victim_inst._detected:
            break

    assert f"[ALERT]{_subj_name}" in victim_inst._detected


# =====================================================================================================================
# Monitor_DonorSvetofor()
# Monitor_CbrKeyRate()
Monitor_ConquestS23_comments()
# Monitor_AcraRaiting_GTLC()



from typing import *


# =====================================================================================================================
class UrlCreator:
    # SETTINGS -------------------------------------
    PROTOCOL: Optional[str] = None
    HOST: Optional[str] = None
    PORT: Optional[int] = None
    ROUTE: Optional[str] = None

    def URL_create(
            self,
            protocol: Optional[str] = None,
            host: Optional[str] = None,
            port: Optional[int] = None,
            route: Optional[str] = None,
    ) -> str:
        if protocol is None:
            protocol = self.PROTOCOL or "http"
        if host is None:
            host = self.HOST or "localhost"
        if port is None:
            port = self.PORT or 80
        if route is None:
            route = self.ROUTE or ""

        if host == "0.0.0.0":
            host = "localhost"

        while route.startswith("/"):
            route = route[1:]

        url = f"{protocol}://{host}:{port}/{route}"
        return url


# =====================================================================================================================



from typing import *

from base_aux.classes.static import TYPE__LAMBDA_CONSTRUCTOR, TYPE__LAMBDA_ARGS, TYPE__LAMBDA_KWARGS
from base_aux.classes.valid_0_aux import ValidAux


# =====================================================================================================================
class Lambda:
    """
    # FIXME: it seems ValidAux have same functions!!! need to combine in one object???
        in this case its perfect separating args without special Ensure_tuple! NO!!!

    GOAL
    ----
    1. delay probable raising Exx on direct execution (use with AttrLambdaCall)
    like creating objects on Cls attributes
        class Cls:
            ATTR = PrivateValues(123)   # Lambda(PrivateValues, 123)

    2. (not serious) replace simple lambda!
    by using lambda you should define args/kwargs any time! and im sick of it!
        func = lambda *args, **kwargs: sum(*args) + sum(**kwargs.values())  # its not a simple lambda!
        func = lambda *args: sum(*args)  # its simple lambda
        result = func(1, 2)
    replace to
        func = Lambda(sum)
        result = func(1, 2)

        func = Lambda(sum, 1, 2)
        result = func()
    its те a good idea to replace lambda fully!
    cause you cant replace following examples
        func_link = lambda source: str(self.Victim(source))
        func_link = lambda source1, source2: self.Victim(source1) == source2


    SPECIALLY CREATED FOR
    ---------------------
    Item for using with AttrLambdaCall

    WHY NOT 1=simple LAMBDA?
    ------------------------
    extremely good point!
    but
    1. in case of at least AttrLambdaCall you cant distinguish method or callable attribute!
    so you explicitly define attributes/objects for later constructions
    and in some point it can be more clear REPLACE LAMBDA by this solvation!!!

    2.

    PARAMS
    ======
    :ivar CONSTRUCTOR: any class or function
    """
    CONSTRUCTOR: TYPE__LAMBDA_CONSTRUCTOR
    ARGS: TYPE__LAMBDA_ARGS = ()
    KWARGS: TYPE__LAMBDA_KWARGS = {}

    # OVERWRITE! ------------------------------------------------------------------------------------------------------
    def __call__(self, *args, **kwargs) -> Any | NoReturn:
        args = args or self.ARGS
        kwargs = kwargs or self.KWARGS
        return self.CONSTRUCTOR(*args, **kwargs)

    def __eq__(self, other) -> bool | NoReturn:
        return ValidAux.compare_doublesided__bool(other, self())

    # UNIVERSAL =======================================================================================================
    def __init__(self, constructor: TYPE__LAMBDA_CONSTRUCTOR, *args, **kwargs) -> None:
        self.CONSTRUCTOR = constructor
        self.ARGS = args
        self.KWARGS = kwargs

    # -----------------------------------------------------------------------------------------------------------------
    def __bool__(self) -> bool | NoReturn:
        return bool(self(*self.ARGS, **self.KWARGS))

    # -----------------------------------------------------------------------------------------------------------------
    def get_result_or_raise(self, *args, **kwargs) -> bool | NoReturn:
        """
        just a direct result
        """
        return self(*args, **kwargs)

    def get_result_or_exx(self, *args, **kwargs) -> bool | Exception:
        """
        SPECIALLY CREATED FOR
        ---------------------
        just in case
        """
        try:
            return self(*args, **kwargs)
        except Exception as exx:
            return exx

    # -----------------------------------------------------------------------------------------------------------------
    def check_raise(self, *args, **kwargs) -> bool:
        """
        SPECIALLY CREATED FOR
        ---------------------
        check Privates in pytest for skipping

        USE LambdaTrySuccess instead!
        """
        try:
            self(*args, **kwargs)
            return False
        except Exception as exx:
            return True

    def check_no_raise(self, *args, **kwargs) -> bool:
        return not self.check_raise(*args, **kwargs)


# =====================================================================================================================
class LambdaBool(Lambda):
    """
    GOAL
    ----
    same as Lambda, in case of get result in bool variant
    +add reverse

    SPECIALLY CREATED FOR
    ---------------------
    classes.Valid.skip_link with Reverse variant

    why Reversing is so important?
    --------------------------------
    because you cant keep callable link and reversing it by simply NOT
        skip_link__direct = bool        # correct
        skip_link__direct = LambdaBool(bool)  # correct
        skip_link__reversal = not bool  # incorrect
        skip_link__reversal = LambdaBool(bool, attr).get_reverse  # correct

    but here we can use lambda
        skip_link__reversal = lambda attr: not bool(attr)  # correct but not so convenient ???

    PARAMS
    ======
    :ivar BOOL_REVERSE: just for LambdaBoolReversed, no need to init
    """

    BOOL_REVERSE: bool = False

    def __call__(self, *args, **kwargs) -> bool | NoReturn:
        args = args or self.ARGS
        kwargs = kwargs or self.KWARGS
        result = bool(self.CONSTRUCTOR(*args, **kwargs))
        if self.BOOL_REVERSE:
            result = not result
        return result

    def get_reverse(self, *args, **kwargs) -> bool | NoReturn:
        """
        if raise - raise

        try not to use in LambdaBoolReversed
        """
        return not self(*args, **kwargs)

    def get_bool_only(self, *args, **kwargs) -> bool:
        """
        if raise - return False, else get result
        """
        try:
            return self(*args, **kwargs)
        except Exception as exx:
            return False

    def get_bool_only__reverse(self, *args, **kwargs) -> bool:
        return not self.get_bool_only(*args, **kwargs)


class LambdaBoolReversed(LambdaBool):
    """
    just a reversed LambdaBool
    """
    BOOL_REVERSE: bool = True


# =====================================================================================================================
class LambdaTrySuccess(LambdaBool):
    """
    just an ability to check if object is not raised on call

    best practice
    -------------
    if LambdaTrySuccess(func):
        return func()

    @pytest.mark.skipif(LambdaTryFail(func), ...)
    """
    def __call__(self, *args, **kwargs) -> bool:
        args = args or self.ARGS
        kwargs = kwargs or self.KWARGS
        try:
            self.CONSTRUCTOR(*args, **kwargs)
            return not self.BOOL_REVERSE
        except:
            return bool(self.BOOL_REVERSE)


class LambdaTryFail(LambdaTrySuccess):
    BOOL_REVERSE: bool = True


# =====================================================================================================================



from typing import *
import sys
import pathlib

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .base import Row_, Dev_, Data_
from .tm import TableModelTemplate
from .th import HeaderViewCB
from .hl import *


# =====================================================================================================================
TYPE__SIZE_TUPLE = tuple[Optional[int], Optional[int]]


# =====================================================================================================================
class Gui(QWidget):
    # SETTINGS --------------------------------------------------
    START: bool = True

    TITLE: str = "[GUI] Template"
    LOGO: str = "logo.jpg"
    CENTER: bool = True

    SIZE_MINIMUM: TYPE__SIZE_TUPLE = (None, None)
    SIZE_MAXIMUM: TYPE__SIZE_TUPLE = (None, None)
    SIZE_FIXED: TYPE__SIZE_TUPLE = (None, None)
    SIZE: TYPE__SIZE_TUPLE = (None, None)
    MOVE: TYPE__SIZE_TUPLE = (None, None)

    FLAGS: dict[Any, str] = {
        # TODO: use separated as CLASS!!! with special FLAG methods!!! sum/del/check/... and try to mark as True/False/None

        # UNCOMMENT IF NEEDED!
        # values are just for information!
        # if some flags overlays/conflict - used last activated

        # BORDER/FRAME/TITLE ------------------------------------------------------------------
        # Qt.FramelessWindowHint: "[frame]hide outer frame/border (with title), to tern back title add WindowTitleHint",
        # Qt.WindowTitleHint: "[title]force turn on (after FramelessWindowHint) window title",
        # Qt.CustomizeWindowHint: "[title]hide std frame border with title",

        # GEOMETRY ----------------------------------------------------------------------------
        # Qt.MSWindowsFixedSizeDialogHint: "[geometry]block window mouse resizing",

        # BTNS --------------------------------------------------------------------------------
        # Qt.WindowSystemMenuHint: "???[btn]deactivate all btns",
        # Qt.WindowMinimizeButtonHint: "[btn]activate ONLY MINimize",
        # Qt.WindowMaximizeButtonHint: "[btn]activate ONLY MAXimize",
        # Qt.WindowMinMaxButtonsHint: "[btn]activate ONLY MAX+MINimize",
        # Qt.WindowCloseButtonHint: "[btn]keep only CLOSE",
        # Qt.WindowContextHelpButtonHint: "[btn]keep only HELP +CLOSE(but inactivated)",

        # LAYERS -------------------------------------------------------------------------------
        # Qt.WindowStaysOnTopHint: "[layer] always on TOP",
        # Qt.WindowStaysOnBottomHint: "[layer] always on BOTTOM",
    }

    # AUXILIARY --------------------------------------------------
    _QAPP: QApplication = QApplication([])

    # COMMON ------------------------------------------------------
    DATA: Optional[Any] = None

    BTN: Optional[QPushButton] = None
    CB: Optional[QCheckBox] = None
    TV: Optional[QTableView] = None
    TM: Optional[QAbstractTableModel] = None
    PTE: Optional[QPlainTextEdit] = None
    HL: Optional[QSyntaxHighlighter] = None
    HL_STYLES: Optional[HlStyles] = None

    def __init__(self, data: Optional[Any] = None):
        super().__init__()

        if data is not None:
            self.DATA = data

        if self.START:
            self.run()

    def run(self):
        self.wgt_create()
        self.slots_connect()

        # GUI SHOW ----------------------------------------------------------------------------------------------------
        self._wgt_main__apply_settings()
        self.show()
        if self.CENTER:
            self._wgt_main__center()

        # starting PYQT in thread - NOT AVAILABLE!!! ---------------------------
        # i tried switch all even show() into it - not working!
        # thread = threading.Thread(target=self.run)
        # thread.start()
        # thread.join()

        exit_code = self._QAPP.exec()
        # time.sleep(5)
        if exit_code == 0:
            print(f"[OK]GUI({exit_code=})closed correctly")
        else:
            print(f"[FAIL]GUI({exit_code=})closed INCORRECTLY")
        sys.exit(exit_code)

    # MAIN WINDOW =====================================================================================================
    def _wgt_main__apply_settings(self) -> None:
        # TITLE --------------------------------------------------
        self.setWindowTitle(self.TITLE)
        self._wgt_main__apply_logo()

        # FLAGS ---------------------------------------------------
        flag_cum = 0
        for flag in self.FLAGS:
            flag_cum |= flag
        if flag_cum:
            self.setWindowFlags(flag_cum)

        # GEOMETRY ------------------------------------------------
        if self.SIZE_MINIMUM[0]:
            self.setMinimumWidth(self.SIZE_MINIMUM[0])
        if self.SIZE_MINIMUM[1]:
            self.setMinimumHeight(self.SIZE_MINIMUM[1])

        if self.SIZE_MAXIMUM[0]:
            self.setMaximumWidth(self.SIZE_MAXIMUM[0])
        if self.SIZE_MAXIMUM[1]:
            self.setMaximumHeight(self.SIZE_MAXIMUM[1])

        if self.SIZE_FIXED[0]:
            self.setFixedWidth(self.SIZE_FIXED[0])
        if self.SIZE_FIXED[1]:
            self.setFixedHeight(self.SIZE_FIXED[1])

        if self.SIZE[0] or self.SIZE[1]:
            width = self.SIZE[0] or self.width()
            height = self.SIZE[1] or self.height()
            self.resize(width, height)

        if self.MOVE[0] or self.MOVE[1]:
            x = self.MOVE[0] or self.x()
            y = self.MOVE[1] or self.y()
            self.move(x, y)

        # USER ------------------------------------------------

    def _wgt_main__apply_logo(self) -> None:
        """
        need square size for logo!
        """
        logo_filepath = pathlib.Path(self.LOGO)
        if logo_filepath.is_file() and logo_filepath.exists():
            self._QAPP.setWindowIcon(QIcon(logo_filepath.name))

            try:
                # turn on logo for python-applications (only for Windows) as associations
                from PyQt5.QtWinExtras import QtWin
                QtWin.setCurrentProcessExplicitAppUserModelID('mycompany.myproduct.subproduct.version')
            except:
                pass

    def _wgt_main__center(self):
        """
        center the main window considering MULTY MONITORS.
        NOTE: work incorrect in INIT!!! use in root module right after wgt.SHOW() not before!!!
        """
        window_geometry = self.frameGeometry()
        # print(f"window_geometry={window_geometry}")      # PyQt5.QtCore.QRect(100, 100, 500, 500)

        display_obj = QApplication.desktop()
        # print(f"display_obj={display_obj}")      # <PyQt5.QtWidgets.QDesktopWidget object at 0x0000020630E771F0>

        display_index = display_obj.screenNumber(display_obj.cursor().pos())
        # print(f"display_index={display_index}")    # 1

        display_geometry = display_obj.screenGeometry(display_index)
        # print(f"display_geometry={display_geometry}")    # PyQt5.QtCore.QRect(1366, 0, 1920, 1080)

        display_central_point = display_geometry.center()
        # print(f"display_central_point={display_central_point}")    # PyQt5.QtCore.QPoint(2325, 539)

        self.move(
            display_central_point.x() - window_geometry.width()//2,
            display_central_point.y() - window_geometry.height()//2
        )

    # WINDOW ==========================================================================================================
    def wgt_create(self) -> None:
        self.BTN_create()
        self.CB_create()
        self.TV_create()
        self.PTE_create()
        self.HL_create()
        self.BTNS_PTE_create()

        # GRID --------------------------------------------------------------------------------------------------------
        layout_grid = QGridLayout()

        # settings -----------------------------------
        layout_grid.setColumnStretch(1, 2)
        layout_grid.setRowStretch(2, 2)

        layout_grid.setHorizontalSpacing(2)
        layout_grid.setVerticalSpacing(1)
        layout_grid.setSpacing(1)

        layout_grid.setColumnMinimumWidth(0, 100)
        layout_grid.setRowMinimumHeight(1, 50)

        # wgts ---------------------------------------
        layout_grid.addWidget(QLabel("00"), 0, 0)
        layout_grid.addWidget(QLabel("01"), 0, 1)
        layout_grid.addWidget(QLabel("02"), 0, 2)
        layout_grid.addWidget(QLabel("03"), 0, 3)

        layout_grid.addWidget(QLabel("10"), 1, 0)
        layout_grid.addWidget(QLabel("11-12"), 1, 1, 1, 2)
        layout_grid.addWidget(QLabel("13"), 1, 3)

        layout_grid.addWidget(QLabel("20"), 2, 0)
        layout_grid.addWidget(QLabel("21-end"), 2, 1, 2, -1)

        # pte_btns -------------------------------------------------------------------------------------------
        # wgts --------------------------------------
        # TODO: FINISH WITH BTNS_BLOCK!!!!

        # layout ------------------------------------
        # layout_h = QHBoxLayout()
        # layout_h.addWidget(self.)

        # layout_main -------------------------------------------------------------------------------------------------
        layout_v = QVBoxLayout()
        layout_v.addLayout(layout_grid)
        layout_v.addWidget(self.CB)
        layout_v.addWidget(self.BTN)
        # layout_v.addLayout(layout_h)
        layout_v.addWidget(self.PTE)

        # layout_main -------------------------------------------------------------------------------------------------
        layout_main = QHBoxLayout()
        layout_main.addWidget(self.TV)
        layout_main.addLayout(layout_v)

        self.setLayout(layout_main)

    # WGTS ============================================================================================================
    def BTN_create(self) -> None:
        self.BTN = QPushButton("BTN")

        # SETTINGS -------------------------
        self.BTN.setText("BTN_mod")
        self.BTN.setToolTip("ToolTip")

        self.BTN.setCheckable(True)
        # self.BTN.setChecked(False)
        # self.BTN.setFlat(True)        # виджет вида Label!!! НО нажатия визуализирует как виджет BTN!!!
        # self.BTN.setDefault(True)     # кажется выделяет рамкой навсегда среди всех?

        # self.BTN.setEnabled(True)
        # self.BTN.setDisabled(True)
        # self.BTN.setVisible(True)
        # self.BTN.setHidden(True)

        # GEOMETRY ----------------------
        # self.BTN.setContentsMargins(1000, 1000, 1000, 1000)     # ничего не дает!!!
        # self.BTN.setSizeIncrement(100, 100)     # не понял!!!

        # self.BTN.setMinimumWidth(5)
        # self.BTN.setMinimumHeight(5)
        # self.BTN.setMinimumSize(5, 5)

        # self.BTN.setMaximumWidth(300)
        # self.BTN.setMaximumHeight(20)
        # self.BTN.setMaximumSize(300, 20)

        # если это сделать, то расширяться не будет!!!!!
        # self.BTN.setFixedWidth(100)      # in pixels
        # self.BTN.setFixedHeight(20)
        # self.BTN.setFixedSize(5, 5)

        # OBJECTS ---------------------------
        # self.BTN.setFont(...)

        # self.BTN.setIcon(...)
        # self.BTN.setIconSize(...)

        # DONT UNDERSTAND -------------------
        # self.BTN.setAutoRepeat(True)
        # self.BTN.setAutoRepeatDelay(2)
        # self.BTN.setAutoRepeatInterval(2)

        # PROPERTIES ------------------------
        # print(self.BTN.isEnabled())
        # print(self.BTN.isCheckable())
        # print(self.BTN.isChecked())
        # print(self.BTN.isDown())
        # print(self.BTN.isFlat())
        # print(self.BTN.isHidden())
        # print(self.BTN.isVisible())
        # print(self.BTN.isDefault())

    def BTNS_PTE_create(self) -> None:
        # TODO: ADD BTNS_BLOCK!!! and apply here!!! used to explore PTE!!!
        # self.BTN_PTE = QPushButton("BTN")
        #
        # # SETTINGS -------------------------
        # self.BTN.setText("BTN_mod")
        # self.BTN.setToolTip("ToolTip")
        #
        # self.BTN.setCheckable(True)
        pass

    def CB_create(self) -> None:
        self.CB = QCheckBox("CB_text")

        # SETTINGS -------------------------
        # self.CB.setText("CB_text")
        self.CB.setText("CB_text")
        self.CB.setTristate()
        # ObjectInfo(self.CB).print()
        """
==========================================================================================
----------OBJECTINFO.PRINT--------------------------------------------------------------------------
str(SOURCE)=<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DEA0>
repr(SOURCE)=<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DEA0>
type(SOURCE)=<class 'PyQt5.QtWidgets.QCheckBox'>
mro(SOURCE)=['QCheckBox', 'QAbstractButton', 'QWidget', 'QObject', 'wrapper', 'QPaintDevice', 'simplewrapper', 'object']
----------SETTINGS----------------------------------------------------------------------------------
self.NAMES__USE_ONLY_PARTS=[]
self.NAMES__SKIP_FULL=[]
self.NAMES__SKIP_PARTS=['init', 'new', 'create', 'enter', 'install', 'set', 'clone', 'copy', 'move', 'next', 'clear', 'reduce', 'close', 'del', 'exit', 'kill', 'exec', 'exec_', 'pyqtConfigure', 'checkout', 'detach', 'run', 'start', 'wait', 'join', 'terminate', 'quit', 'disconnect', 'pop', 'popleft', 'append', 'appendleft', 'extend', 'extendleft', 'add', 'insert', 'reverse', 'rotate', 'sort']
self.HIDE_BUILD_IN=None
self.LOG_ITER=None
self.MAX_LINE_LEN=100
self.MAX_ITER_ITEMS=5
----------log_iter(wait last touched)---------------------------------------------------------------
----------SKIPPED_FULLNAMES-------------------------------------------------------------------------
----------SKIPPED_PARTNAMES-------------------------------------------------------------------------
1:	__delattr__
2:	__init__
3:	__init_subclass__
4:	__new__
5:	__reduce__
6:	__reduce_ex__
7:	__setattr__
8:	addAction
9:	addActions
10:	clearFocus
11:	clearMask
12:	close
13:	closeEvent
14:	create
15:	createWindowContainer
16:	deleteLater
17:	disconnect
18:	disconnectNotify
19:	enterEvent
20:	initPainter
21:	initStyleOption
22:	insertAction
23:	insertActions
24:	installEventFilter
25:	killTimer
26:	move
27:	moveEvent
28:	moveToThread
29:	nextCheckState
30:	nextInFocusChain
31:	pyqtConfigure
32:	removeAction
33:	removeEventFilter
34:	setAcceptDrops
35:	setAccessibleDescription
36:	setAccessibleName
37:	setAttribute
38:	setAutoExclusive
39:	setAutoFillBackground
40:	setAutoRepeat
41:	setAutoRepeatDelay
42:	setAutoRepeatInterval
43:	setBackgroundRole
44:	setBaseSize
45:	setCheckState
46:	setCheckable
47:	setChecked
48:	setContentsMargins
49:	setContextMenuPolicy
50:	setCursor
51:	setDisabled
52:	setDown
53:	setEnabled
54:	setFixedHeight
55:	setFixedSize
56:	setFixedWidth
57:	setFocus
58:	setFocusPolicy
59:	setFocusProxy
60:	setFont
61:	setForegroundRole
62:	setGeometry
63:	setGraphicsEffect
64:	setHidden
65:	setIcon
66:	setIconSize
67:	setInputMethodHints
68:	setLayout
69:	setLayoutDirection
70:	setLocale
71:	setMask
72:	setMaximumHeight
73:	setMaximumSize
74:	setMaximumWidth
75:	setMinimumHeight
76:	setMinimumSize
77:	setMinimumWidth
78:	setMouseTracking
79:	setObjectName
80:	setPalette
81:	setParent
82:	setProperty
83:	setShortcut
84:	setShortcutAutoRepeat
85:	setShortcutEnabled
86:	setSizeIncrement
87:	setSizePolicy
88:	setStatusTip
89:	setStyle
90:	setStyleSheet
91:	setTabOrder
92:	setTabletTracking
93:	setText
94:	setToolTip
95:	setToolTipDuration
96:	setTristate
97:	setUpdatesEnabled
98:	setVisible
99:	setWhatsThis
100:	setWindowFilePath
101:	setWindowFlag
102:	setWindowFlags
103:	setWindowIcon
104:	setWindowIconText
105:	setWindowModality
106:	setWindowModified
107:	setWindowOpacity
108:	setWindowRole
109:	setWindowState
110:	setWindowTitle
111:	startTimer
112:	unsetCursor
113:	unsetLayoutDirection
114:	unsetLocale
----------PROPERTIES__ELEMENTARY_SINGLE-------------------------------------------------------------
DrawChildren        	RenderFlag  :2
DrawWindowBackground	RenderFlag  :1
IgnoreMask          	RenderFlag  :4
PdmDepth            	PaintDeviceMetric:6
PdmDevicePixelRatio 	PaintDeviceMetric:11
PdmDevicePixelRatioScaled	PaintDeviceMetric:12
PdmDpiX             	PaintDeviceMetric:7
PdmDpiY             	PaintDeviceMetric:8
PdmHeight           	PaintDeviceMetric:2
PdmHeightMM         	PaintDeviceMetric:4
PdmNumColors        	PaintDeviceMetric:5
PdmPhysicalDpiX     	PaintDeviceMetric:9
PdmPhysicalDpiY     	PaintDeviceMetric:10
PdmWidth            	PaintDeviceMetric:1
PdmWidthMM          	PaintDeviceMetric:3
__doc__             	str         :QCheckBox(parent: Optional[QWidget] = None)
QCheckBox(text: ...
__module__          	str         :PyQt5.QtWidgets
__weakref__         	NoneType    :None
----------PROPERTIES__ELEMENTARY_COLLECTION---------------------------------------------------------
__dict__            	dict        :{}
----------PROPERTIES__OBJECTS-----------------------------------------------------------------------
staticMetaObject    	QMetaObject :<PyQt5.QtCore.QMetaObject object at 0x000001A3759622D0>
----------PROPERTIES__EXX---------------------------------------------------------------------------
----------METHODS__ELEMENTARY_SINGLE----------------------------------------------------------------
PaintDeviceMetric   	PaintDeviceMetric:0
RenderFlag          	RenderFlag  :0
__getstate__        	NoneType    :None
__hash__            	int         :112597741034
__repr__            	str         :<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DEA0>
__sizeof__          	int         :120
__str__             	str         :<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DEA0>
acceptDrops         	bool        :False
accessibleDescription	str         :
accessibleName      	str         :
activateWindow      	NoneType    :None
adjustSize          	NoneType    :None
animateClick        	NoneType    :None
autoExclusive       	bool        :False
autoFillBackground  	bool        :False
autoRepeat          	bool        :False
autoRepeatDelay     	int         :300
autoRepeatInterval  	int         :100
backgroundRole      	ColorRole   :1
checkState          	CheckState  :0
checkStateSet       	NoneType    :None
click               	NoneType    :None
colorCount          	int         :256
contextMenuPolicy   	ContextMenuPolicy:1
depth               	int         :32
destroy             	NoneType    :None
devType             	int         :1
devicePixelRatio    	int         :1
devicePixelRatioF   	float       :1.0
devicePixelRatioFScale	float       :65536.0
dumpObjectInfo      	NoneType    :None
dumpObjectTree      	NoneType    :None
effectiveWinId      	NoneType    :None
ensurePolished      	NoneType    :None
focusNextChild      	bool        :False
focusPolicy         	FocusPolicy :11
focusPreviousChild  	bool        :False
focusProxy          	NoneType    :None
foregroundRole      	ColorRole   :0
grabKeyboard        	NoneType    :None
grabMouse           	NoneType    :None
graphicsEffect      	NoneType    :None
graphicsProxyWidget 	NoneType    :None
group               	NoneType    :None
hasFocus            	bool        :False
hasHeightForWidth   	bool        :False
hasMouseTracking    	bool        :True
hasTabletTracking   	bool        :False
height              	int         :17
heightMM            	int         :3
hide                	NoneType    :None
isActiveWindow      	bool        :False
isCheckable         	bool        :True
isChecked           	bool        :True
isDown              	bool        :False
isEnabled           	bool        :True
isFullScreen        	bool        :False
isHidden            	bool        :True
isLeftToRight       	bool        :True
isMaximized         	bool        :False
isMinimized         	bool        :False
isModal             	bool        :False
isRightToLeft       	bool        :False
isTristate          	bool        :False
isVisible           	bool        :False
isWidgetType        	bool        :True
isWindow            	bool        :True
isWindowModified    	bool        :False
isWindowType        	bool        :False
layout              	NoneType    :None
layoutDirection     	LayoutDirection:0
logicalDpiX         	int         :96
logicalDpiY         	int         :96
lower               	NoneType    :None
maximumHeight       	int         :16777215
maximumWidth        	int         :16777215
minimumHeight       	int         :0
minimumWidth        	int         :0
nativeParentWidget  	NoneType    :None
objectName          	str         :
paintEngine         	NoneType    :None
paintingActive      	bool        :False
parent              	NoneType    :None
parentWidget        	NoneType    :None
physicalDpiX        	int         :109
physicalDpiY        	int         :109
raise_              	NoneType    :None
releaseKeyboard     	NoneType    :None
releaseMouse        	NoneType    :None
repaint             	NoneType    :None
sender              	NoneType    :None
senderSignalIndex   	int         :-1
sharedPainter       	NoneType    :None
show                	NoneType    :None
showFullScreen      	NoneType    :None
showMaximized       	NoneType    :None
showMinimized       	NoneType    :None
showNormal          	NoneType    :None
signalsBlocked      	bool        :False
statusTip           	str         :
styleSheet          	str         :
text                	str         :CB_text
toggle              	NoneType    :None
toolTip             	str         :
toolTipDuration     	int         :-1
underMouse          	bool        :False
update              	NoneType    :None
updateGeometry      	NoneType    :None
updateMicroFocus    	NoneType    :None
updatesEnabled      	bool        :True
whatsThis           	str         :
width               	int         :2560
widthMM             	int         :597
windowFilePath      	str         :
windowIconText      	str         :
windowModality      	WindowModality:0
windowOpacity       	float       :1.0
windowRole          	str         :
windowTitle         	str         :
windowType          	WindowType  :1
x                   	int         :2559
y                   	int         :-7
----------METHODS__ELEMENTARY_COLLECTION------------------------------------------------------------
__dir__             	list        :['__module__', '__doc__', 'actionEvent', 'changeEvent', 'che...
                    	str         :	__module__
                    	str         :	__doc__
                    	str         :	actionEvent
                    	str         :	changeEvent
                    	str         :	checkState
                    	            :	...
actions             	list        :[]
children            	list        :[]
dynamicPropertyNames	list        :[]
getContentsMargins  	tuple       :(0, 0, 0, 0)
----------METHODS__OBJECTS--------------------------------------------------------------------------
RenderFlags         	RenderFlags :<PyQt5.QtWidgets.QWidget.RenderFlags object at 0x000001A375960D60>
__class__           	QCheckBox   :<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DFC0>
__subclasshook__    	NotImplementedType:NotImplemented
baseSize            	QSize       :PyQt5.QtCore.QSize()
childrenRect        	QRect       :PyQt5.QtCore.QRect()
childrenRegion      	QRegion     :<PyQt5.QtGui.QRegion object at 0x000001A375961460>
contentsMargins     	QMargins    :<PyQt5.QtCore.QMargins object at 0x000001A375961540>
contentsRect        	QRect       :PyQt5.QtCore.QRect(0, 0, 62, 17)
cursor              	QCursor     :<PyQt5.QtGui.QCursor object at 0x000001A375961620>
focusWidget         	QCheckBox   :<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DEA0>
font                	QFont       :<PyQt5.QtGui.QFont object at 0x000001A3759618C0>
fontInfo            	QFontInfo   :<PyQt5.QtGui.QFontInfo object at 0x000001A375961930>
fontMetrics         	QFontMetrics:<PyQt5.QtGui.QFontMetrics object at 0x000001A3759619A0>
frameGeometry       	QRect       :PyQt5.QtCore.QRect(0, 0, 61, 16)
frameSize           	QSize       :PyQt5.QtCore.QSize(61, 16)
geometry            	QRect       :PyQt5.QtCore.QRect(0, 0, 62, 17)
grab                	QPixmap     :<PyQt5.QtGui.QPixmap object at 0x000001A375961B60>
icon                	QIcon       :<PyQt5.QtGui.QIcon object at 0x000001A37592E050>
iconSize            	QSize       :PyQt5.QtCore.QSize(16, 16)
inputMethodHints    	InputMethodHints:<PyQt5.QtCore.Qt.InputMethodHints object at 0x000001A375...
keyboardGrabber     	QCheckBox   :<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DEA0>
locale              	QLocale     :<PyQt5.QtCore.QLocale object at 0x000001A375961FC0>
mask                	QRegion     :<PyQt5.QtGui.QRegion object at 0x000001A3759621F0>
maximumSize         	QSize       :PyQt5.QtCore.QSize(16777215, 16777215)
metaObject          	QMetaObject :<PyQt5.QtCore.QMetaObject object at 0x000001A3759622D0>
minimumSize         	QSize       :PyQt5.QtCore.QSize()
minimumSizeHint     	QSize       :PyQt5.QtCore.QSize(62, 17)
mouseGrabber        	QCheckBox   :<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DEA0>
normalGeometry      	QRect       :PyQt5.QtCore.QRect(0, 0, 62, 17)
palette             	QPalette    :<PyQt5.QtGui.QPalette object at 0x000001A375962490>
pos                 	QPoint      :PyQt5.QtCore.QPoint()
previousInFocusChain	QCheckBox   :<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DEA0>
rect                	QRect       :PyQt5.QtCore.QRect(0, 0, 62, 17)
saveGeometry        	QByteArray  :b'\x01\xd9\xd0\xcb\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00...
screen              	QScreen     :<PyQt5.QtGui.QScreen object at 0x000001A37592E0E0>
shortcut            	QKeySequence:<PyQt5.QtGui.QKeySequence object at 0x000001A3759627A0>
size                	QSize       :PyQt5.QtCore.QSize(2560, 1377)
sizeHint            	QSize       :PyQt5.QtCore.QSize(62, 17)
sizeIncrement       	QSize       :PyQt5.QtCore.QSize()
sizePolicy          	QSizePolicy :<PyQt5.QtWidgets.QSizePolicy object at 0x000001A3759629D0>
style               	QCommonStyle:<PyQt5.QtWidgets.QCommonStyle object at 0x000001A37592E170>
thread              	QThread     :<PyQt5.QtCore.QThread object at 0x000001A37592E200>
visibleRegion       	QRegion     :<PyQt5.QtGui.QRegion object at 0x000001A375962B90>
winId               	voidptr     :<sip.voidptr object at 0x000001A3759FE9D0>
window              	QCheckBox   :<PyQt5.QtWidgets.QCheckBox object at 0x000001A37592DEA0>
windowFlags         	WindowFlags :<PyQt5.QtCore.Qt.WindowFlags object at 0x000001A375962C00>
windowHandle        	QWindow     :<PyQt5.QtGui.QWindow object at 0x000001A37592E290>
windowIcon          	QIcon       :<PyQt5.QtGui.QIcon object at 0x000001A37592E320>
windowState         	WindowStates:<PyQt5.QtCore.Qt.WindowStates object at 0x000001A375962C70>
----------METHODS__EXX------------------------------------------------------------------------------
__eq__              	TypeError   :TypeError('expected 1 argument, got 0')
__format__          	TypeError   :TypeError('QCheckBox.__format__() takes exactly one argument...
__ge__              	TypeError   :TypeError('expected 1 argument, got 0')
__getattr__         	TypeError   :TypeError('__getattr__(self, name: Optional[str]): not enoug...
__getattribute__    	TypeError   :TypeError('expected 1 argument, got 0')
__gt__              	TypeError   :TypeError('expected 1 argument, got 0')
__le__              	TypeError   :TypeError('expected 1 argument, got 0')
__lt__              	TypeError   :TypeError('expected 1 argument, got 0')
__ne__              	TypeError   :TypeError('expected 1 argument, got 0')
actionEvent         	TypeError   :TypeError('actionEvent(self, a0: Optional[QActionEvent]): no...
blockSignals        	TypeError   :TypeError('blockSignals(self, b: bool): not enough arguments')
changeEvent         	TypeError   :TypeError('changeEvent(self, e: Optional[QEvent]): not enoug...
childAt             	TypeError   :TypeError('arguments did not match any overloaded call:\n  c...
childEvent          	TypeError   :TypeError('childEvent(self, a0: Optional[QChildEvent]): not ...
clicked             	TypeError   :TypeError('native Qt signal is not callable')
connectNotify       	TypeError   :TypeError('connectNotify(self, signal: QMetaMethod): not eno...
contextMenuEvent    	TypeError   :TypeError('contextMenuEvent(self, a0: Optional[QContextMenuE...
customContextMenuRequested	TypeError   :TypeError('native Qt signal is not callable')
customEvent         	TypeError   :TypeError('customEvent(self, a0: Optional[QEvent]): not enou...
destroyed           	TypeError   :TypeError('native Qt signal is not callable')
dragEnterEvent      	TypeError   :TypeError('dragEnterEvent(self, a0: Optional[QDragEnterEvent...
dragLeaveEvent      	TypeError   :TypeError('dragLeaveEvent(self, a0: Optional[QDragLeaveEvent...
dragMoveEvent       	TypeError   :TypeError('dragMoveEvent(self, a0: Optional[QDragMoveEvent])...
dropEvent           	TypeError   :TypeError('dropEvent(self, a0: Optional[QDropEvent]): not en...
event               	TypeError   :TypeError('event(self, e: Optional[QEvent]): not enough argu...
eventFilter         	TypeError   :TypeError('eventFilter(self, a0: Optional[QObject], a1: Opti...
find                	TypeError   :TypeError('find(a0: PyQt5.sip.voidptr): not enough arguments')
findChild           	TypeError   :TypeError("arguments did not match any overloaded call:\n  f...
findChildren        	TypeError   :TypeError("arguments did not match any overloaded call:\n  f...
focusInEvent        	TypeError   :TypeError('focusInEvent(self, e: Optional[QFocusEvent]): not...
focusNextPrevChild  	TypeError   :TypeError('focusNextPrevChild(self, next: bool): not enough ...
focusOutEvent       	TypeError   :TypeError('focusOutEvent(self, e: Optional[QFocusEvent]): no...
grabGesture         	TypeError   :TypeError('grabGesture(self, type: Qt.GestureType, flags: Un...
grabShortcut        	TypeError   :TypeError('grabShortcut(self, key: Union[QKeySequence, QKeyS...
heightForWidth      	TypeError   :TypeError('heightForWidth(self, a0: int): not enough arguments')
hideEvent           	TypeError   :TypeError('hideEvent(self, a0: Optional[QHideEvent]): not en...
hitButton           	TypeError   :TypeError('hitButton(self, pos: QPoint): not enough arguments')
inherits            	TypeError   :TypeError('inherits(self, classname: Optional[str]): not eno...
inputMethodEvent    	TypeError   :TypeError('inputMethodEvent(self, a0: Optional[QInputMethodE...
inputMethodQuery    	TypeError   :TypeError('inputMethodQuery(self, a0: Qt.InputMethodQuery): ...
isAncestorOf        	TypeError   :TypeError('isAncestorOf(self, child: Optional[QWidget]): not...
isEnabledTo         	TypeError   :TypeError('isEnabledTo(self, a0: Optional[QWidget]): not eno...
isSignalConnected   	TypeError   :TypeError('isSignalConnected(self, signal: QMetaMethod): not...
isVisibleTo         	TypeError   :TypeError('isVisibleTo(self, a0: Optional[QWidget]): not eno...
keyPressEvent       	TypeError   :TypeError('keyPressEvent(self, e: Optional[QKeyEvent]): not ...
keyReleaseEvent     	TypeError   :TypeError('keyReleaseEvent(self, e: Optional[QKeyEvent]): no...
leaveEvent          	TypeError   :TypeError('leaveEvent(self, a0: Optional[QEvent]): not enoug...
mapFrom             	TypeError   :TypeError('mapFrom(self, a0: Optional[QWidget], a1: QPoint):...
mapFromGlobal       	TypeError   :TypeError('mapFromGlobal(self, a0: QPoint): not enough arguments')
mapFromParent       	TypeError   :TypeError('mapFromParent(self, a0: QPoint): not enough arguments')
mapTo               	TypeError   :TypeError('mapTo(self, a0: Optional[QWidget], a1: QPoint): n...
mapToGlobal         	TypeError   :TypeError('mapToGlobal(self, a0: QPoint): not enough arguments')
mapToParent         	TypeError   :TypeError('mapToParent(self, a0: QPoint): not enough arguments')
metric              	TypeError   :TypeError('metric(self, a0: QPaintDevice.PaintDeviceMetric):...
mouseDoubleClickEvent	TypeError   :TypeError('mouseDoubleClickEvent(self, a0: Optional[QMouseE...
mouseMoveEvent      	TypeError   :TypeError('mouseMoveEvent(self, a0: Optional[QMouseEvent]): ...
mousePressEvent     	TypeError   :TypeError('mousePressEvent(self, e: Optional[QMouseEvent]): ...
mouseReleaseEvent   	TypeError   :TypeError('mouseReleaseEvent(self, e: Optional[QMouseEvent])...
nativeEvent         	TypeError   :TypeError('nativeEvent(self, eventType: Union[QByteArray, by...
objectNameChanged   	TypeError   :TypeError('native Qt signal is not callable')
overrideWindowFlags 	TypeError   :TypeError('overrideWindowFlags(self, type: Union[Qt.WindowFl...
overrideWindowState 	TypeError   :TypeError('overrideWindowState(self, state: Union[Qt.WindowS...
paintEvent          	TypeError   :TypeError('paintEvent(self, a0: Optional[QPaintEvent]): not ...
pressed             	TypeError   :TypeError('native Qt signal is not callable')
property            	TypeError   :TypeError('property(self, name: Optional[str]): not enough a...
receivers           	TypeError   :TypeError('receivers(self, signal: PYQT_SIGNAL): not enough ...
releaseShortcut     	TypeError   :TypeError('releaseShortcut(self, id: int): not enough arguments')
released            	TypeError   :TypeError('native Qt signal is not callable')
render              	TypeError   :TypeError('arguments did not match any overloaded call:\n  r...
resize              	TypeError   :TypeError('arguments did not match any overloaded call:\n  r...
resizeEvent         	TypeError   :TypeError('resizeEvent(self, a0: Optional[QResizeEvent]): no...
restoreGeometry     	TypeError   :TypeError('restoreGeometry(self, geometry: Union[QByteArray,...
scroll              	TypeError   :TypeError('arguments did not match any overloaded call:\n  s...
showEvent           	TypeError   :TypeError('showEvent(self, a0: Optional[QShowEvent]): not en...
stackUnder          	TypeError   :TypeError('stackUnder(self, a0: Optional[QWidget]): not enou...
stateChanged        	TypeError   :TypeError('native Qt signal is not callable')
tabletEvent         	TypeError   :TypeError('tabletEvent(self, a0: Optional[QTabletEvent]): no...
testAttribute       	TypeError   :TypeError('testAttribute(self, attribute: Qt.WidgetAttribute...
timerEvent          	TypeError   :TypeError('timerEvent(self, e: Optional[QTimerEvent]): not e...
toggled             	TypeError   :TypeError('native Qt signal is not callable')
tr                  	TypeError   :TypeError('tr(self, sourceText: Optional[str], disambiguatio...
ungrabGesture       	TypeError   :TypeError('ungrabGesture(self, type: Qt.GestureType): not en...
wheelEvent          	TypeError   :TypeError('wheelEvent(self, a0: Optional[QWheelEvent]): not ...
windowIconChanged   	TypeError   :TypeError('native Qt signal is not callable')
windowIconTextChanged	TypeError   :TypeError('native Qt signal is not callable')
windowTitleChanged  	TypeError   :TypeError('native Qt signal is not callable')
==========================================================================================
        """

    def TV_create(self) -> None:
        # PREPARE ------------------------
        if self.DATA is None:
            self.DATA = Data_([Row_(f"row{index}") for index in range(5)], [Dev_(f"dev{index}") for index in range(4)])

        self.TM = TableModelTemplate(self.DATA)

        # WORK ---------------------------
        self.TV = QTableView()
        self.TV.setModel(self.TM)

        # STYLE -----
        # self.TV.setShowGrid(True)
        # self.TV.setFont(QFont("Calibri (Body)", 12))  # сразу на все!!! и на заголоски и на ячейки
        # self.TV.setStyleSheet("gridline-color: rgb(255, 0, 0)")

        # self.TV.setSortingEnabled(True)     # enable sorting

        # HEADER ---------
        self.TV.setHorizontalHeader(HeaderViewCB(self.DATA))   # you can add some additional HV object!

        # hh = self.TV.horizontalHeader()
        # hh.setStretchLastSection(True)
        # hh.setSectionsClickable(False)
        # hh.setSectionsMovable(False)
        # hh.setVisible(False)
        # hh.swapSections(1,2)

        # GEOMETRY  ----------
        # self.TV.setMinimumSize(400, 300)
        # self.TV.resize(400, 300)
        # self.TV.setColumnWidth(0, 100)
        self.TV.resizeColumnsToContents()   # set column width to fit contents - NEED AFTER SetHEADER!!!

    def PTE_create(self) -> None:
        self.PTE = QPlainTextEdit()

        # METHODS ORIGINAL ---------------------------------
        # self.PTE.setEnabled(True)
        # self.PTE.setUndoRedoEnabled(True)
        # self.PTE.setReadOnly(True)
        # self.PTE.setMaximumBlockCount(15)

        # self.PTE.clear()
        self.PTE.setPlainText("setPlainText")
        self.PTE.appendPlainText("appendPlainText")
        self.PTE.appendPlainText("result=None")
        self.PTE.appendPlainText("result=True")
        self.PTE.appendPlainText("result=False")
        # self.PTE.appendHtml("")
        # self.PTE.anchorAt(#)
        # self.PTE.setSizeAdjustPolicy(#)

        # METHODS COMMON -----------------------------------
        self.PTE.setFont(QFont("Calibri (Body)", 7))

        # ObjectInfo(self.PTE).print()

        # BODY
        # body: str = self.PTE.toPlainText()

    def HL_create(self) -> None:
        if self.HL_STYLES and self.PTE:
            self.HL = Highlighter(document=self.PTE.document(), styles=self.HL_STYLES)

    # SLOTS ===========================================================================================================
    def slots_connect(self) -> None:
        if self.BTN:
            self.BTN.clicked.connect(self._wgt_main__center)
            if self.BTN.isCheckable():
                self.BTN.toggled.connect(self.BTN__toggled)
            else:
                self.BTN.clicked.connect(self.BTN__toggled)

        if self.CB:
            self.CB.stateChanged.connect(self.CB__changed)

        if self.TV:
            self.TV.selectionModel().selectionChanged.connect(self.TV_selectionChanged)
            # self.TV.horizontalHeader().sectionClicked.connect(self.TV_hh_sectionClicked)

    def BTN__toggled(self, state: Optional[bool] = None) -> None:
        print(f"BTN__toggled {state=}")

    def CB__changed(self, state: Optional[int] = None) -> None:
        """
        :param state:
            0 - unchecked
            1 - halfCHecked (only if isTristate)
            2 - checked (even if not isTristate)
        """
        print(f"CB__changed {state=}")

    def TV_selectionChanged(self, first: QItemSelection, last: QItemSelection) -> None:
        # print("selectionChanged")
        # print(f"{first=}")  # first=<PyQt5.QtCore.QItemSelection object at 0x000001C79A107460>
        # ObjectInfo(first.indexes()[0]).print(_log_iter=True, skip_fullnames=["takeFirst", "takeLast"])

        # print(f"{first=}")

        if not first:
            print(f"selected first NotSelectable Index {first=}")
            return

        try:
            index: QModelIndex = first.indexes()[0]
        except:
            return

        row = index.row()
        col = index.column()

        self.PTE.setPlainText(f"{row=}/{col=}")

    # EVENTS ==========================================================================================================
    pass    # events list see in source code!
    pass    # events list see in source code!
    pass    # events list see in source code!
    pass    # events list see in source code!
    pass    # events list see in source code!

    # def mouseMoveEvent(self, a0: typing.Optional[QtGui.QMouseEvent]) -> None: ...

    def moveEvent(self,  a0: Optional[QMoveEvent] = None, QMoveEvent: Any = None) -> None:
        # print(self.geometry().x(), self.geometry().y())
        pass

    def resizeEvent(self, a0: Optional[QResizeEvent] = None, ResizeEvent: Any = None) -> None:
        # print(self.size())
        pass

    # mouse POINTER -------------------------------------------------
    def enterEvent(self, a0: Optional[QEvent] = None) -> None:
        """mouse get aria over the wgt
        """
        # print("mouse enterEvent")
        pass

    def leaveEvent(self, a0: Optional[QEvent] = None) -> None:
        """mouse leave aria over the wgt
        """
        # print("mouse leaveEvent")
        pass

    # mouse CLICK -------------------------------------------------
    def mousePressEvent(self, a0: Optional[QMouseEvent] = None) -> None:
        """will not apply when click on wgt btns!"""
        # print("mouse mousePressEvent")
        pass

    def mouseDoubleClickEvent(self, a0: Optional[QMouseEvent] = None) -> None:
        """will not apply when click on wgt btns!
        DIFFERENCE between DoubleClick and PressEvent
            if detected DoubleClick it will generate DoubleClickEvent else PressEvent
                mouse mousePressEvent
                mouse mouseReleaseEvent
                mouse mouseDoubleClickEvent
                mouse mouseReleaseEvent
        """
        # print("mouse mouseDoubleClickEvent")
        pass

    def mouseReleaseEvent(self, a0: Optional[QMouseEvent] = None) -> None:
        """
        work always! both for One/Double click!
            mouse mousePressEvent
            mouse mouseReleaseEvent
            mouse mouseDoubleClickEvent
            mouse mouseReleaseEvent
        """
        # print("mouse mouseReleaseEvent")
        pass

    # NOT WORKING ------------------------------------------
    # def focusOutEvent(self, a0: Optional[QEvent]) -> None:
    #     """mouse leave aria over the wgt
    #     """
    #     # print("focusOutEvent")
    #     pass
    #
    # def focusInEvent(self, a0: Optional[QEvent]) -> None:
    #     """mouse leave aria over the wgt
    #     """
    #     # print("focusInEvent")
    #     pass


# =====================================================================================================================



from base_aux.stock import *

import time
from typing import *
import datetime as dt
import re
import json
import threading

import MetaTrader5 as mt5

import pandas as pd
import numpy as np

from base_aux.privates import *
from base_aux.alerts import *


# =====================================================================================================================
# TODO: ADD STABILITY STRESSFUL! like in IMAP!!!! connect if lost!


# =====================================================================================================================
Type__Symbol = Union[str, mt5.SymbolInfo]
Type__SymbolOpt = Optional[Type__Symbol]

Type__Tf = int
Type__TfOpt = Optional[Type__Tf]

Type__PdSeries = pd.core.series.Series
Type__IndicatorValues = Union[None, float, Type__PdSeries]


# =====================================================================================================================
class Exx__Mt5Auth(Exception):
    pass


class Exx__Mt5SymbolName(Exception):
    pass


# =====================================================================================================================
class PrivateMT5(PrivateAuthAuto):
    SECTION = "AUTH_MT5_DEF"
    SERVER: str


# =====================================================================================================================
class MT5:
    MT5_AUTH: PrivateMT5 = PrivateMT5()

    SYMBOL: Type__Symbol = "BRX3"
    TF: int = mt5.TIMEFRAME_M10
    __MT5_SYMBOLS_AVAILABLE: List[mt5.SymbolInfo] = None

    BAR_LAST: np.ndarray = None
    """
    time
    (1675118400, 85.41, 85.43, 85.21, 85.21, 225, 1, 1065)

    bar["time"] --> 1675118400
    """
    TICK_LAST: mt5.Tick = None
    """
    self.BAR_LAST_TICK=Tick(time=1675468684, bid=83.45, ask=83.51, last=83.5, volume=6, time_msc=1675468684950, flags=30, volume_real=6.0)
    type(self.BAR_LAST_TICK)=<class 'Tick'>
    """

    bar__lost: bool = None        # DONT SET! used only in STRATEGY!!!
    tick__lost: bool = None

    _symbols_volume_price: Dict[str, float] = {}     # used as

    # =================================================================================================================
    def __init__(self, _tf: Optional[int] = None, _symbol: Optional[str] = None):
        super().__init__()

        self.TF = _tf or self.TF
        self.SYMBOL = _symbol or self.SYMBOL

        self.mt5_connect()
        self._SYMBOL_init(exx_if_none=False)

    def __del__(self):
        mt5.shutdown()

    # CONNECT ---------------------------------------------------------------------------------------------------------
    def mt5_connect(self) -> Optional[NoReturn]:
        result = mt5.initialize(login=int(self.MT5_AUTH.USER), password=self.MT5_AUTH.PWD, server=self.MT5_AUTH.SERVER)
        msg = f"[{result}]initialize[{mt5.last_error()=}]"
        print(msg)
        if not result:
            msg += f"БЫВАЕТ СЛЕТАЕТ ПАРОЛЬ!!! становится неверным - нужно просто обновить его в MT5"
            msg += f"\n{self.MT5_AUTH}"

            print(msg)
            raise Exx__Mt5Auth(msg)

    # SYMBOL ==========================================================================================================
    def _SYMBOL_init(self, exx_if_none: bool = True) -> Optional[NoReturn]:
        self.SYMBOL = self.SYMBOL__get_active(exx_if_none=exx_if_none)

    def SYMBOL__get_active(self, _symbol: Type__SymbolOpt = None, exx_if_none: bool = True) -> Union[mt5.SymbolInfo, NoReturn]:
        _symbol = _symbol or self.SYMBOL
        if isinstance(_symbol, str):
            _symbol = mt5.symbol_info(_symbol)
            last_error = mt5.last_error()
            if last_error[0] != 1:
                msg = f"incorrect {_symbol=}/{last_error=}"
                raise Exx__Mt5SymbolName(msg)

        if not isinstance(_symbol, mt5.SymbolInfo):
            msg = f"incorrect {_symbol=}"
            if exx_if_none:
                raise Exx__Mt5SymbolName(msg)

        return _symbol

    def TF__get_active(self, _tf: Type__TfOpt = None):
        return _tf or self.TF

    # AVAILABLE -------------------------------------------------------------------------------------------------------
    @property
    def _MT5_SYMBOLS_AVAILABLE(self) -> List[mt5.SymbolInfo]:
        """
        too much time to get all items! dont use it without special needs!
        """
        if not self.__MT5_SYMBOLS_AVAILABLE:
            self.__MT5_SYMBOLS_AVAILABLE = mt5.symbols_get()
        return self.__MT5_SYMBOLS_AVAILABLE

    def _mt5_symbols_available__find(
            self,
            mask: str = "",
            only_rus: bool = False
    ) -> List[mt5.SymbolInfo]:
        # count=12976 ALL!!!!
        # count=275 rus!!!!
        count = 0
        result: List[mt5.SymbolInfo] = []

        if not mask:
            symbols = self._MT5_SYMBOLS_AVAILABLE
        else:
            symbols = mt5.symbols_get(mask)

        print("*"*100)
        for item in symbols:
            # FILTER LONG NAMES like for Options
            if len(item.name) > 5:
                continue

            # FILTER RUS
            if only_rus:
                if any([
                    item.isin != "moex.stock",
                    not re.match(pattern=r"[а-яА-Я]", string=item.description),
                    item.name.startswith("RU00"),
                    # item.currency_base != "RUS",  # always OK!
                    # "-RM" in item.name,     #count=12301
                ]):
                    continue

            # GET
            count += 1
            result.append(item)
            print(item.name)

        print("*"*100)
        print(f"result={[item.name for item in result]}")
        print(f"{count=}")
        print("*"*100)

        return result

    # SHOW ------------------------------------------------------------------------------------------------------------
    def _mt5_symbol_show(self, show: bool = True, _symbol: Type__SymbolOpt = None) -> bool:
        _symbol = self.SYMBOL__get_active(_symbol)
        result = mt5.symbol_select(_symbol.name, show)
        print(f"[{result}]_mt5_symbol_show({_symbol})={mt5.last_error()=}")
        return result

    def _mt5_symbol_show__check(self, _symbol: Type__SymbolOpt = None) -> bool:
        _symbol = self.SYMBOL__get_active(_symbol)
        if _symbol:
            return _symbol.select

    # INFO ------------------------------------------------------------------------------------------------------------
    def _symbols_info__print_compare(self, symbols: List[Type__Symbol] = ["SBER", "AAPL-RM", "PYPL-RM"]):
        """
        since SYMBOL_NAME not added into chart gui list - it will return zero to many attributes!

        ****************************************************************************************************
        INSTRUMENTS                   =['SBER', 'AAPL-RM', 'PYPL-RM']
        ****************************************************************************************************
        custom                        =[False, False, False]
        chart_mode                    =[1, 1, 1]
        select                        =[True, False, False]
        visible                       =[True, False, False]
        session_deals                 =[0, 0, 0]
        session_buy_orders            =[0, 0, 0]
        session_sell_orders           =[0, 0, 0]
        volume                        =[34, 0, 0]
        volumehigh                    =[12811, 0, 0]
        volumelow                     =[1, 0, 0]
        time                          =[1671839399, 0, 0]
        digits                        =[2, 0, 0]
        spread                        =[6, 0, 0]
        spread_float                  =[True, True, True]
        ticks_bookdepth               =[32, 32, 32]
        trade_calc_mode               =[32, 32, 32]
        trade_mode                    =[4, 4, 4]
        start_time                    =[0, 0, 0]
        expiration_time               =[0, 0, 0]
        trade_stops_level             =[0, 0, 0]
        trade_freeze_level            =[0, 0, 0]
        trade_exemode                 =[3, 3, 3]
        swap_mode                     =[0, 0, 0]
        swap_rollover3days            =[3, 3, 3]
        margin_hedged_use_leg         =[False, False, False]
        expiration_mode               =[15, 15, 15]
        filling_mode                  =[3, 3, 3]
        order_mode                    =[63, 63, 63]
        order_gtc_mode                =[2, 2, 2]
        option_mode                   =[0, 0, 0]
        option_right                  =[0, 0, 0]
        bid                           =[137.85, 0.0, 0.0]
        bidhigh                       =[139.01, 0.0, 0.0]
        bidlow                        =[136.81, 0.0, 0.0]
        ask                           =[137.91, 0.0, 0.0]
        askhigh                       =[138.36, 0.0, 0.0]
        asklow                        =[136.82, 0.0, 0.0]
        last                          =[137.94, 0.0, 0.0]
        lasthigh                      =[138.26, 0.0, 0.0]
        lastlow                       =[136.81, 0.0, 0.0]
        volume_real                   =[34.0, 0.0, 0.0]
        volumehigh_real               =[12811.0, 0.0, 0.0]
        volumelow_real                =[1.0, 0.0, 0.0]
        option_strike                 =[0.0, 0.0, 0.0]
        point                         =[0.01, 1.0, 1.0]
        trade_tick_value              =[0.1, 1.0, 1.0]
        trade_tick_value_profit       =[0.1, 1.0, 1.0]
        trade_tick_value_loss         =[0.1, 1.0, 1.0]
        trade_tick_size               =[0.01, 1.0, 1.0]
        trade_contract_size           =[10.0, 1.0, 1.0]
        trade_accrued_interest        =[0.0, 0.0, 0.0]
        trade_face_value              =[0.0, 0.0, 0.0]
        trade_liquidity_rate          =[1.0, 1.0, 1.0]
        volume_min                    =[1.0, 1.0, 1.0]
        volume_max                    =[100000000.0, 100000000.0, 100000000.0]
        volume_step                   =[1.0, 1.0, 1.0]
        volume_limit                  =[0.0, 0.0, 0.0]
        swap_long                     =[0.0, 0.0, 0.0]
        swap_short                    =[0.0, 0.0, 0.0]
        margin_initial                =[0.0, 0.0, 0.0]
        margin_maintenance            =[0.0, 0.0, 0.0]
        session_volume                =[0.0, 0.0, 0.0]
        session_turnover              =[0.0, 0.0, 0.0]
        session_interest              =[0.0, 0.0, 0.0]
        session_buy_orders_volume     =[0.0, 0.0, 0.0]
        session_sell_orders_volume    =[0.0, 0.0, 0.0]
        session_open                  =[137.49, 0.0, 0.0]
        session_close                 =[137.69, 0.0, 0.0]
        session_aw                    =[0.0, 0.0, 0.0]
        session_price_settlement      =[0.0, 0.0, 0.0]
        session_price_limit_min       =[0.0, 0.0, 0.0]
        session_price_limit_max       =[0.0, 0.0, 0.0]
        margin_hedged                 =[0.0, 0.0, 0.0]
        price_change                  =[0.1816, 0.0, 0.0]
        price_volatility              =[0.0, 0.0, 0.0]
        price_theoretical             =[0.0, 0.0, 0.0]
        price_greeks_delta            =[0.0, 0.0, 0.0]
        price_greeks_theta            =[0.0, 0.0, 0.0]
        price_greeks_gamma            =[0.0, 0.0, 0.0]
        price_greeks_vega             =[0.0, 0.0, 0.0]
        price_greeks_rho              =[0.0, 0.0, 0.0]
        price_greeks_omega            =[0.0, 0.0, 0.0]
        price_sensitivity             =[0.0, 0.0, 0.0]
        basis                         =['', '', '']
        category                      =['', '', '']
        currency_base                 =['RUR', 'RUR', 'RUR']
        currency_profit               =['RUR', 'RUR', 'RUR']
        currency_margin               =['RUR', 'RUR', 'RUR']
        bank                          =['', '', '']
        description                   =['Сбербанк России ПАО ао', 'Apple Inc.', 'PayPal Holdings, Inc.']
        exchange                      =['', '', '']
        formula                       =['', '', '']
        isin                          =['moex.stock', 'moex.stock', 'moex.stock']
        name                          =['SBER', 'AAPL-RM', 'PYPL-RM']
        page                          =['', '', '']
        path                          =['MOEX\\SBER', 'MOEX\\AAPL-RM', 'MOEX\\PYPL-RM']
        ****************************************************************************************************
        """
        items = []
        for symbol in list(symbols):
            item = self.SYMBOL__get_active(symbol)
            if item:
                items.append(item._asdict())
            else:
                symbols.remove(symbol)

        if not items:
            return

        print("*"*100)
        key = "INSTRUMENTS"
        print(f"{key:30}={symbols}")
        print("*"*100)
        for key in items[0]:
            value = []
            for item in items:
                value.append(item.get(key))
            print(f"{key:30}={value}")

        print("*"*100)

    # VOLUME_PRICE -----------------------------------------------------
    def _symbol_volume_price_get__last_day_finished(self, _symbol: Type__SymbolOpt = None, _devider: Optional[int] = None) -> float:
        """
        VolumePrice as priceMean * Volume
        +save result into self.symbols_volume_price for threading usage!

        value will be differ from official becouse of i get mean price!
        https://www.moex.com/ru/marketdata/?g=4#/mode=groups&group=4&collection=3&boardgroup=57&data_type=current&category=main

        :param _symbol:
        :return:
        """
        _devider = _devider or 1000 * 1000
        bar = self.bars_get__count(_symbol=_symbol, _tf=mt5.TIMEFRAME_D1)
        # print(f"{bar['real_volume']=}")

        item = mt5.symbol_info(_symbol)
        contracts_per_lot = item.trade_contract_size
        contracts = contracts_per_lot * bar["real_volume"]
        volume_price = contracts * (bar["high"] + bar["low"])/2
        # print(f"{volume_price=}")

        result = round(volume_price[0]/_devider)
        self._symbols_volume_price.update({_symbol: result})
        return result

    def _symbols_get_sorted_volume_price(self, limit_min=None, limit_max=None, _symbols: Optional[List[str]] = None, _devider: Optional[int] = None) -> Dict[str, float]:
        """

        (400 * 1000 * 1000)
        ['SBER', 'GAZP', 'LKOH', 'PLZL', 'MGNT', 'NVTK', 'UWGN', 'LQDT', 'VTBR', 'GMKN', 'LSNG', 'SNGS', 'ROSN', 'CHMF']

        :param limit_min:
        :param limit_max:
        :return:
        """
        _devider = _devider or 1000 * 1000
        limit_min = limit_min if limit_min is not None else 100 * 1000 * 1000 / _devider
        _symbols = _symbols or Symbols.SYMBOLS__RUS_FINAM

        # LOAD ---------------------------------------------------
        for symbol in _symbols:
            threading.Thread(target=self._symbol_volume_price_get__last_day_finished, kwargs={"_symbol": symbol, "_devider": _devider}).start()

        while threading.active_count() > 1:
            time.sleep(1)

        # FILTER ---------------------------------------------------
        for symbol, value in dict(self._symbols_volume_price).items():
            if limit_max and limit_max < value:
                self._symbols_volume_price.pop(symbol)
            if limit_min > value:
                self._symbols_volume_price.pop(symbol)

        # SORT -----------------------------------------------------
        self._symbols_volume_price = dict(sorted(self._symbols_volume_price.items(), key=lambda x: x[1], reverse=True))

        # PRINT ----------------------------------------------------
        result_pretty = json.dumps(self._symbols_volume_price, indent=4)
        print(result_pretty)
        return self._symbols_volume_price

    # BAR HISTORY =====================================================================================================
    pass

    # BAR -------------------------------------------------------------------------------------------------------------
    def bar_new__wait(self, sleep: int = 10) -> None:
        # TODO: use BarTime to resolve LOST!!!
        counter = 0
        while not self.bar_last__update():
            counter += 1
            self.bar__lost = True
            msg = f"[WARN]bar lost [{counter=}]"
            print(msg)
            time.sleep(sleep)
        self.bar__lost = False

    def bar_last__update(self) -> Optional[True]:
        bar_new = self.bars_get__count(count=1)
        if not self.BAR_LAST or bar_new != self.BAR_LAST:
            self.BAR_LAST = bar_new
            return True

    def tick_last__update(self, _symbol: Type__SymbolOpt = None, wait_tick_load: bool = True) -> bool:
        """

        SYMBOL_NAME have to be in terminal! otherwise error
            [False]tick_last__update()=mt5.last_error()=(-4, 'Terminal: Not found')
        """
        _symbol = self.SYMBOL__get_active(_symbol)
        result = False
        while True:
            tick = mt5.symbol_info_tick(self.SYMBOL)
            result = tick != self.TICK_LAST
            if result:
                break

            if not wait_tick_load:
                break
            time.sleep(1)

        if result:
            print(f"update[{self.TICK_LAST=}]{_symbol}/{mt5.last_error()=}")
            # Tick(time=1665770358, bid=62.437, ask=63.312, last=0.0, volume=0, time_msc=1665770358179, flags=6, volume_real=0.0)
            self.TICK_LAST = tick
        return result

    # HISTORY ---------------------------------------------------------------------------------------------------------
    def bars__check_actual(
            self,
            _symbol: Type__SymbolOpt = None,
            _tf: Type__TfOpt = None
    ) -> bool:
        _symbol = self.SYMBOL__get_active(_symbol)
        _tf_td = dt.timedelta(minutes=self.TF__get_active(_tf))

        result = False
        last = self.TICK_LAST
        if last:
            last_dt = dt.datetime.fromtimestamp(last.time)
            result = (last_dt + _tf_td) >= dt.datetime.today()
        return result

    def bars_get__count(
            self,
            count: int = 1,
            tf_split: Optional[int] = None,
            shrink: Optional[bool] = None,
            _start: Optional[int] = None,
            _symbol: Type__SymbolOpt = None,
            _tf: Type__TfOpt = None
    ) -> Union[np.ndarray]:
        """get history bars

        :param count: get exact count of bars
        :param tf_split: correct count of bars in case of using split tf
        :param shrink: True - if you need in results full correct split bars (for ADX you need new close/open)
            Otherwise, dont use it! for RSI you need only close, which will be accessed directly by steps

        :_start: 0 is actual not finished!
        ['time', 'open', 'high', 'low', 'close', 'tick_volume', 'spread', 'real_volume']
            [(1695763800, 93.83, 93.88, 93.78, 93.88, 172, 1, 723)]
            elem=1695763800/<class 'numpy.int64'>
            elem=93.83/<class 'numpy.float64'>
            elem=93.88/<class 'numpy.float64'>
            elem=93.78/<class 'numpy.float64'>
            elem=93.88/<class 'numpy.float64'>
            elem=172/<class 'numpy.uint64'>
            elem=1/<class 'numpy.intc'>
            elem=723/<class 'numpy.uint64'>
        """
        _symbol = self.SYMBOL__get_active(_symbol)
        _tf = self.TF__get_active(_tf)
        tf_split = tf_split or 1
        _start = _start or 1

        bars = mt5.copy_rates_from_pos(_symbol.name, _tf, _start, count * tf_split)
        # if not bars:
        #     print(f"{_symbol=}/{bars=}")
        #     return

        # for bar in bars:
        #     print(f"{type(bar)}={bar}")     # <class 'numpy.void'>=(1671753600, 137.49, 138.26, 136.81, 137.94, 53823, 0, 2283422)

        if tf_split > 1 and shrink:
            bars = HistoryShifted_Shrink(bars, tf_split).shrink()

        # if count == 1:
        #     # bars = [(1695729000, 92.3, 92.42, 92.22, 92.23, 944, 1, 3381)]
        #     return bars[0]  # numpy.void
        # else:
        #     # bars = [(1695728400, 92.16, 92.32, 92.1, 92.31, 578, 1, 2764)
        #     #         (1695729000, 92.3, 92.42, 92.22, 92.23, 944, 1, 3381)]
        #     return bars     # numpy.ndarray

        return bars

    # INDICATOR =======================================================================================================
    def _indicator_get_by_obj(
            self,
            indicator_params: IndicatorParamsBase,
            *,
            return_tail: Optional[int] = 1,
            tf_split: Optional[int] = None,

            _bars: Optional[np.ndarray] = None,
            _add_history: Optional[int] = None,
            _tf: Type__TfOpt = None,
            _symbol: Type__SymbolOpt = None,
    ) -> Type__IndicatorValues:
        # GET -----------------------------
        bars_np = _bars or self.bars_get__count(
            count=indicator_params.bars_expected__get(),
            tf_split=tf_split,
            shrink=indicator_params.NAME == IndicatorName.ADX,
            _start=_add_history + 1,
            _symbol=_symbol,
            _tf=_tf
        )

        # DF -----------------------------
        df = pd.DataFrame(bars_np)

        # ACTUAL INDICATOR -----------------------------
        params_dict = indicator_params.params_dict__get()
        if indicator_params.NAME == IndicatorName.WMA:
            # df = ta.wma(**df, **params_dict)
            df = df.ta.wma(**params_dict)
        elif indicator_params.NAME == IndicatorName.STOCH:
            df = df.ta.stoch(**params_dict)
            # df = df.ta.stoch(**{"fask_k": 10, "slow_d": 5})
            # df = df.ta.stoch(fast_k=10, slow_k=2, slow_d=2)
        elif indicator_params.NAME == IndicatorName.ADX:
            df = df.ta.adx(**params_dict)
        elif indicator_params.NAME == IndicatorName.MACD:
            df = df.ta.macd(**params_dict)
        else:
            msg = f"cant detect name [{indicator_params.NAME=}]"
            raise Exception(msg)

        # ROUND -----------------------------
        df = df.iloc[:].round(indicator_params.ROUND)

        # FINAL -----------------------------
        name = indicator_params.column_name__get()
        try:        # FIXME: check directly?
            # if result gives only one column - its not have header! so it will raise!
            # like WMA!
            # but it used for others! like ADX/STOCH/MACD!
            df = df[name]
        except:
            pass

        if return_tail == 1:
            result = df.iloc[len(df) - 1]
        elif not return_tail:
            result = df
        else:
            try:
                # if less then need!
                result = df[-return_tail::]
            except:
                pass

        return result

    def indicator_WMA(self, args: Collection[int], **kwargs) -> Type__IndicatorValues:
        return self._indicator_get_by_obj(IndicatorParams_WMA(*args), **kwargs)

    def indicator_STOCH(self, args: Collection[int], **kwargs) -> Type__IndicatorValues:
        return self._indicator_get_by_obj(IndicatorParams_STOCH(*args), **kwargs)

    def indicator_ADX(self, args: Collection[int], **kwargs) -> Type__IndicatorValues:
        return self._indicator_get_by_obj(IndicatorParams_ADX(*args), **kwargs)

    def indicator_MACD(self, args: Collection[int], **kwargs) -> Type__IndicatorValues:
        return self._indicator_get_by_obj(IndicatorParams_MACD(*args), **kwargs)


# =====================================================================================================================



from base_aux.stock import *

import sys
from typing import *
import threading
import abc

import pandas as pd
import numpy as np
import MetaTrader5 as mt5

from base_aux.funcs import *
from base_aux.alerts import *
from base_aux.threads import *


# =====================================================================================================================
pass    # settings
np.set_printoptions(threshold=sys.maxsize, linewidth=300)


# =====================================================================================================================
class MonitorBase(MT5, abc.ABC, threading.Thread):
    """Base monitor class.

    :ivar ALERT: Alert class used instantiating send msg
    :ivar counter: count active monitoring steps
    :ivar state_calculating: state when step is in active calculating phase
        new bar get + calculatings started and not yet finished
    :ivar LOAD_HISTORY_BARS: apply calculations on exact history bars
    """
    ALERT: Type[AlertBase] = None
    counter: int = 0
    state_calculating: bool = None
    LOAD_HISTORY_BARS: int = 10

    def __init__(self):
        super().__init__()
        self.ALERT(f"start strategy")
        self.LOAD_HISTORY_BARS = self.LOAD_HISTORY_BARS or 0
        self.start()

    def run(self):
        """manage cycles of monitoring
        """
        while True:
            self.counter += 1
            self.state_calculating = True

            print()
            msg = f"[OK]cycle new [{self.counter=}]"
            print(msg)

            self.run_strategy_cycle_one()
            self.ALERT.threads_wait_all()

            self.state_calculating = False

            if not self.LOAD_HISTORY_BARS:
                self.bar_new__wait()
            else:
                self.LOAD_HISTORY_BARS -= 1

    @abc.abstractmethod
    def run_strategy_cycle_one(self):
        """Execute one full cycle
        """
        pass


# =====================================================================================================================
class AlertTradeADX(AlertSelect.TELEGRAM_DEF):
    """exact alert object for ADX Strategy
    """
    pass


class ThreadManagerADX(ThreadsManager):
    """Manager which create new group of threads
    """
    pass


class MonitorADX(MonitorBase):
    TF = 1
    ALERT: Type[AlertBase] = AlertTradeADX

    # LOCAL --------------------------
    ADX_FAST: Tuple[int, int] = (6, 2)
    ADX_SLOW: Tuple[int, int] = (10, 7)

    PULSE_TAIL: int = 20
    THRESH_ADX_FAST_FULL: int = 80  # 80
    THRESH_ADX_SLOW_FULL: int = 50  # 50

    state_full_column_1_10: bool = None

    TF_SHIFTED_MAX: int = 120
    LOAD_HISTORY_BARS: int = 1
    RESULTS: np.ndarray = None

    ARRAY_INTERPRETER: Dict[Literal[-1, 0, 1, 2], Any] = {
        -1: " ",    # not exists
        0: "_",     # exists and no Pulse
        # 1: "*",     # Pulse
        # 2: "#"      # edge pulse
        3: "#"      # edge pulse
    }

    def __init__(self):
        self.RESULTS = np.full(shape=(self.TF_SHIFTED_MAX, self.TF_SHIFTED_MAX), fill_value=-1, dtype=int)
        super().__init__()

    def run_strategy_cycle_one(self):
        for tf_shifted in range(1, self.TF_SHIFTED_MAX + 1):
            self.calculate__tf_shifted(tf_shifted)
        ThreadManagerADX().wait_all()
        self.adx_full_pulse_1_10_check()
        self.print__array_interpreted()
        # self.ALERT("tf_shifted checked")

    def adx_full_pulse_1_10_check(self) -> None:
        results_10 = self.results_length_get(10)
        results_15 = self.results_length_get(15)
        counter = 0
        if (results_10 > 0).sum() or (results_15 > 0).sum():
            for tf_shifted in range(1, 10):
                if (self.results_length_get(tf_shifted) > 0).sum():
                    counter += 1

        if counter >= 8:
            if not self.state_full_column_1_10:
                self.state_full_column_1_10 = True
                self.ALERT(body=f"ADX FULL COLUMN!")
        else:
            self.state_full_column_1_10 = False

    def print__array_interpreted(self) -> None:
        text = array_2d_get_compact_str(
            array=self.RESULTS,
            interpreter=self.ARRAY_INTERPRETER,
            separate_rows=20,
            wrap=True,
            use_rows_num=True
        )
        print()
        print(text)
        print()
        print()
        print()
        return

    @ThreadManagerADX().decorator__to_thread
    def calculate__tf_shifted(self, tf_shifted: int) -> None:
        tf_shifted_index = tf_shifted - 1
        # DATA ----------------------------------------
        adx_fast = self.indicator_ADX(self.ADX_FAST, tf_split=tf_shifted, return_tail=self.PULSE_TAIL, _add_history=self.LOAD_HISTORY_BARS)
        adx_slow = self.indicator_ADX(self.ADX_SLOW, tf_split=tf_shifted, return_tail=self.PULSE_TAIL, _add_history=self.LOAD_HISTORY_BARS)

        adx_fast.reset_index(drop=True, inplace=True)
        adx_slow.reset_index(drop=True, inplace=True)

        adx_fast_slow = pd.concat([adx_fast, adx_slow], axis=1)

        # CHECKS ----------------------------------------
        # state-1=PULSE IN CURRENT STEP
        state_pulse__tf_shifted = self.check__state_pulse__tf_shifted(adx_fast, adx_slow)
        result__tf_shifted = int(state_pulse__tf_shifted)   # 0/1

        # state-2=FIRST PULSE IN CURRENT STEP
        state_pulse__tf_shifted_edge = self.check__state_pulse__tf_shifted_edge(adx_fast, adx_slow)
        if state_pulse__tf_shifted_edge:
            result__tf_shifted = 2

        self.RESULTS[tf_shifted_index] = np.roll(self.RESULTS[tf_shifted_index], -1)
        self.RESULTS[tf_shifted_index][-1] = result__tf_shifted

        # state-3=FIRST PULSE IN GROUP STEPS
        results_steps = self.results_length_get(tf_shifted)
        count_steps__existed = (results_steps >= 0).sum()
        count_steps__pulsed = (results_steps > 0).sum()

        state_pulse__group_edge = False
        if state_pulse__tf_shifted_edge:
            if count_steps__existed > 1:
                state_pulse__group_edge = count_steps__pulsed == 1
            elif count_steps__existed == 1:
                state_pulse__group_edge = True

        if state_pulse__group_edge:
            result__tf_shifted = 3
        self.RESULTS[tf_shifted_index][-1] = result__tf_shifted
        results_steps = self.RESULTS[tf_shifted_index][self.TF_SHIFTED_MAX - tf_shifted: self.TF_SHIFTED_MAX]

        if result__tf_shifted == 3 and not self.LOAD_HISTORY_BARS:
            msg = f"[m{self.TF}/sh={tf_shifted}]\n"
            msg += f"{adx_fast_slow}\n"
            msg += f"{list(results_steps)}\n"
            self.ALERT(body=msg)

    def results_length_get(self, length: int):
        return self.RESULTS[length - 1][self.TF_SHIFTED_MAX - length: self.TF_SHIFTED_MAX]

    def check__state_pulse__tf_shifted(self, adx_fast: Type__PdSeries, adx_slow: Type__PdSeries) -> bool:
        return np.max(adx_fast) >= self.THRESH_ADX_FAST_FULL and np.max(adx_slow) >= self.THRESH_ADX_SLOW_FULL

    def check__state_pulse__tf_shifted_edge(self, adx_fast: Type__PdSeries, adx_slow: Type__PdSeries) -> bool:
        if np.max(adx_fast) >= self.THRESH_ADX_FAST_FULL:
            if np.max(adx_slow.head(len(adx_slow) - 1)) < self.THRESH_ADX_SLOW_FULL <= adx_slow.iloc[len(adx_slow) - 1]:
                return True

        if np.max(adx_slow) >= self.THRESH_ADX_SLOW_FULL:
            if np.max(adx_fast.head(len(adx_fast) - 1)) < self.THRESH_ADX_FAST_FULL <= adx_fast.iloc[len(adx_fast) - 1]:
                return True

        return False


# =====================================================================================================================
pass     # ============================================================================================================
pass     # ============================================================================================================
pass     # ============================================================================================================
pass     # ============================================================================================================
pass     # ============================================================================================================


# =====================================================================================================================
class Alert_MapDrawer(AlertSelect.TELEGRAM_DEF):
    """exact alert object for ADX Strategy
    """
    pass


class ThreadManager_MapDrawer_Tf(ThreadsManager):
    """Manager which create new group of threads
    """
    pass


class ThreadManager_MapDrawer_Shift(ThreadsManager):
    """Manager which create new group of threads
    """
    pass


class IndicatorMapDrawer_Simple(MT5, abc.ABC, threading.Thread):
    ALERT: Type[AlertBase] = Alert_MapDrawer
    INDICATOR: Type[IndicatorParamsBase] = IndicatorParams_RSI
    INDICATOR_SETTINGS: Iterable[int] = (5, )

    TF = mt5.TIMEFRAME_D1

    MAP_LENGTH: int = 10    #100
    MAP_HEIGHT: int = 1    #10

    MAP: np.ndarray = None

    history: np.ndarray = None

    def __init__(self):
        super().__init__()
        self.ALERT(f"start strategy")
        self.start()

    def run(self):
        self.calculate_tfs()
        print()
        print(self.MAP)
        print()

    def calculate_tfs(self):
        """create exactly the MAP = all rows
        """

        # TODO: MOVE INTO MT5!!! in indicator!
        for tf_assembled in range(1, self.MAP_HEIGHT + 1):
            self.calculate_tf(tf_assembled)
        ThreadManager_MapDrawer_Tf().wait_all()

    @ThreadManager_MapDrawer_Tf().decorator__to_thread
    def calculate_tf(self, tf_split: int):
        """calculates all steps for exact TF - its just one row in MAP!
        """
        # define final count of bars
        count_need = self.INDICATOR(*self.INDICATOR_SETTINGS).bars_expected__get() + self.MAP_LENGTH

        # get history
        history = self.bars_get__count(count=count_need, tf_split=tf_split)

        # add results
        for shift in range(tf_split):
            self.calculate_shift(shift, tf_split, history)
        ThreadManager_MapDrawer_Shift().wait_all()

        # results = collect
        threads_sorted = sorted(ThreadManager_MapDrawer_Shift().THREAD_ITEMS, key=lambda x: x.args[0])
        results: np.ndarray = np.array([thread.result for thread in threads_sorted])

        # results = flat
        result = results.flatten("F")

        # save row to MAP
        self.MAP[tf_split - 1] = result

    @ThreadManager_MapDrawer_Shift().decorator__to_thread
    def calculate_shift(self, shift: int, tf_split: int, history: np.ndarray) -> np.ndarray:
        history = history[shift::tf_split]
        result = self._indicator_get_by_obj(self.INDICATOR(*self.INDICATOR_SETTINGS), _bars=history, return_tail=self.MAP_LENGTH//tf_split)
        print()
        print(result)
        print()
        return result


# =====================================================================================================================



from typing import *
from threading import Lock


# =====================================================================================================================
class Exx_SingletonNestingLevels(Exception):
    """Exception when used several unsuitable levels in nesting!

    EXAMPLE:
        VictimBase = SingletonWMetaCall
        setattr(VictimBase, "attr", 0)
        class Victim1(VictimBase):
            attr = 1

        assert VictimBase().attr == 0
        try:
            assert Victim1().attr == 1
        except Exx_SingletonDifferentNestingLevels:
            pass
        else:
            assert False

    MAIN RULES:
    1. always instantiate only last Classes in your tree project!
    """
    pass


# =====================================================================================================================
class SingletonManagerBase:
    """Base (manager) for classes to create singletons.

    GOALS:
    1. threading mutex
    2. prevent and raise incorrect singleton usage

    :ivar _SINGLETONS: collection of created singletons instances.
        when you create several classes, you maybe need to keep access to all of them.
        Used in tests!
        Dont access via instance! recommended using via classname!
    :ivar _MUTEX_SINGLETON: mutex for safe creating items
    :ivar _CLS_BLOCKED: set of classes which can NOT be used as singleton! for correct singletons reason!
        if we instantiate one class and then instantiate one of other class nesting this one -
        before you did not understand your architecture mistake, but now it will rase!
    :ivar _CLS_USED: set of classes which already used as singleton!
    :ivar __INSTANCE: actual singleton instance for current class, here it is just for showing exact usage and correct visualising in IDE!

    Bytheway, you can directly access to all this collections from any your singleton instance or even class! its basic!
    """
    _SINGLETONS: List['SingletonManagerBase'] = []
    _CLS_BLOCKED: Set[Any] = set()
    _CLS_USED: Set[Any] = set()

    _MUTEX_SINGLETON: Lock = Lock()  # keep this special name! dont use just MUTEX! (need uniq! in case of exists in source!)
    __INSTANCE: Any     # just to define usage!

    @classmethod
    def _check_correct_instantiating_singletons(cls, cls_obj: Any) -> Optional[NoReturn]:
        """realise checking right instantiating singletons in project

        :param cls_obj: the original class object!
            (this is very important using it here! using access to direct CLS is incorrect for meta!)
        """
        # check blocked --------------------------------------------
        if cls_obj in cls._CLS_BLOCKED:
            msg = f"{cls_obj.__name__=} WAS BLOCKED before by creating singleton in upper nesting level"
            for cls_blocked in cls._CLS_BLOCKED:
                msg += f"\n\t{cls_blocked.__name__=}"
            raise Exx_SingletonNestingLevels(msg)

        cls._CLS_USED.add(cls_obj)

        for cls_mro in cls_obj.__mro__[1:]:
            if cls_mro in cls._CLS_USED:
                msg = f"{cls_obj.__name__=} WAS USED before by creating singleton in less nesting level"
                for cls_used in cls._CLS_USED:
                    msg += f"\n\t{cls_used.__name__=}"
                raise Exx_SingletonNestingLevels(msg)
            cls._CLS_BLOCKED.add(cls_mro)

    @classmethod
    def _drop_all(cls) -> None:
        """delete all singletons!

        created just for correct testing
        """
        while cls._SINGLETONS:
            try:
                delattr(cls._SINGLETONS.pop(), "__INSTANCE")
            except:
                pass

    @classmethod
    def instance__check_created_and_enshure_it(cls, cls_obj: Any) -> bool:
        """check if instance not created yet and create it

        :returns: result of checking was it exists or not
        """
        result = hasattr(cls_obj, '__INSTANCE')
        if not result:
            setattr(cls_obj, '__INSTANCE', None)
        return result

    @classmethod
    def instance__collect(cls, cls_obj: Any) -> None:
        """collect all singleton objects from all classes

        dont know why but it is not working! both in meta and noMeta
        """
        if cls_obj.__INSTANCE not in cls_obj._SINGLETONS:
            cls_obj._SINGLETONS.append(cls_obj.__INSTANCE)

    # @classmethod
    # def _instance_del(cls) -> None:
    #     """Delete class instance!
    #
    #     expected using for tests!
    #     """
    #     try:
    #         delattr(cls, "__INSTANCE")
    #     except:
    #         pass


# =====================================================================================================================
class SingletonMetaCallClass(SingletonManagerBase, type):
    """metaclass which create the singletons by Call method

    USAGE
    -----
        class MySingleton(metaclass=_SingletonMeta):
            pass
    """
    def __call__(cls, *args, **kwargs):
        cls._check_correct_instantiating_singletons(cls)     # dont place into mutex!

        # -----------------------------------------
        cls._MUTEX_SINGLETON.acquire()

        if not cls.instance__check_created_and_enshure_it(cls):
            cls.__INSTANCE = super().__call__(*args, **kwargs)

            # singleton_group_class = cls.__mro__[1]
            # if not hasattr(singleton_group_class, '_INSTANCES'):
            #     setattr(singleton_group_class, '_INSTANCES', [])
            #     singleton_group_class._INSTANCES = []
            # singleton_group_class._INSTANCES.append(cls.__INSTANCE)

        # ------------------------
        # cls.instance__collect(cls)    # here it is not working!
        if cls.__INSTANCE not in cls._SINGLETONS:
            cls._SINGLETONS.append(cls.__INSTANCE)
        # ------------------------

        cls._MUTEX_SINGLETON.release()
        # -----------------------------------------
        return cls.__INSTANCE


class SingletonByCallMeta(metaclass=SingletonMetaCallClass):
    """same as original metaclass SingletonMetaCallClass but just an another variant of using it by simple nesting
    (without direct using metaclass parameter).

    USAGE
    -----
    BEST PRACTICE
    Use only one level nesting!
        class Victim1(Singleton):
            pass
        class Victim2(Singleton):
            pass
        class VictimINCORRECT(Victim1):
            pass
    """


# =====================================================================================================================
class SingletonByNew(SingletonManagerBase):
    """Singleton manager, creating them by using NEW method without metaclassing
    else one variant after SingletonWMetaCall, in case of metaclass is not acceptable.

    USEFUL CASES:
    1. you need to use some metaclass (cant set two metaclasses).
    2. (main difference) always reinit instance on instantiating! (would be called __init__(*args/kwargs)!
        if blank init - will be the same as SingletonWMetaCall
    """
    def __new__(cls, *args, **kwargs):
        cls._check_correct_instantiating_singletons(cls)     # dont place into mutex!

        # -----------------------------------------
        cls._MUTEX_SINGLETON.acquire()
        if not cls.instance__check_created_and_enshure_it(cls):
            cls.__INSTANCE = super().__new__(cls)

        # ------------------------
        # cls.instance__collect(cls)    # here it is not working!
        if cls.__INSTANCE not in cls._SINGLETONS:
            cls._SINGLETONS.append(cls.__INSTANCE)
        # ------------------------

        cls._MUTEX_SINGLETON.release()
        # -----------------------------------------
        return cls.__INSTANCE


# =====================================================================================================================



from typing import *
# from base_aux.funcs import Iterables
from base_aux.funcs import IterAux, Explicit
from base_aux.classes import *
from base_aux.objects import ObjectInfo


# =====================================================================================================================
class DictCaseinsense(dict):
    """
    just a Caseinsense dict
    """
    # __getattr__ = dict.get
    # __setattr__ = dict.__setitem__
    # __delattr__ = dict.__delitem__
    # __iter__ = dict.__iter__
    # __copy__ = dict.copy

    # __repr__ = dict.__repr__  # и так работает!
    # __str__ = dict.__str__    # и так работает!
    # __len__ = dict.__len__    # и так работает!

    # -----------------------------------------------------------------------------------------------------------------
    def _getitem_original(self, item: Any) -> Explicit | None:
        return IterAux.item__get_original__case_insensitive(item, self)

    def pop(self, item: Any) -> None:
        item_original = self._getitem_original(item)
        if item_original is None:
            item_original = item
        else:
            item_original = item_original()

        super().pop(item_original)

    def get(self, item: Any) -> None:
        key_original = self._getitem_original(item)
        if key_original is None:
            return
        else:
            key_original = key_original()  # EXPLICIT get value

        return super().get(key_original)

    # def set(self, item: Any, value: Any) -> None:

    def update(self, m, /, **kwargs) -> None:
        for item, value in m.items():
            key_original = self._getitem_original(item)
            if key_original is None:
                key_original = item
            else:
                key_original = key_original()  # EXPLICIT get value

            super().update({key_original: value})

    def __contains__(self, item: Any) -> bool:
        if self._getitem_original(item) is not None:
            return True

    # -----------------------------------------------------------------------------------------------------------------
    # ITEM is universal!
    def __getitem__(self, item: Any) -> Any | NoReturn:
        return self.get(item)

    def __setitem__(self, item: Any, value: Any) -> None | NoReturn:
        self.update({item: value})

    def __delitem__(self, item: Any) -> None:
        self.pop(item)


# =====================================================================================================================
class DictDots(DictCaseinsense):
    """
    dot.notation access to dictionary keys.
    RULES
    ------
    1. Levels
        first level - WriteRead
        all other nested levels - ReadOnly for root level! but no errors when set new!
    2. caseInsensitive
    3. when object created it create new copy Dict - so it not the same as dict link!




    # FIXME: below is old!
    CANT CREATE SINGLETON!
    cant create singleton!!! i was trying!!! even by typical def __***

    typical usage:
        my_dict = {'param1': 'GOOD'}
        nested_dict = {'param1': 'nested GOOD too'}
        print(my_dict)  # {'param1': 'GOOD'}

        # EXECUTE
        my_dict = DictDotAttrAccess(my_dict)

        # READING
        print(my_dict)  # {'param1': 'GOOD'}
        print(my_dict.param1)  # 'GOOD'
        print(my_dict["param1"])  # 'GOOD'

        # CHANGE
        my_dict.param1 = 123
        print(my_dict)  # {'param1': 123}
        print(my_dict.param1)  # 123
        my_dict["param1"] = 1000
        print(my_dict)  # {'param1': 1000}
        print(my_dict.param1)  # 1000

        # NESTED
        my_dict.nested = nested_dict
        print(my_dict)  # {'param1': 1000, 'nested': {'param1': 'nested GOOD too'}}
        # print(my_dict.nested.param1)  # AttributeError: 'dict' object has no attribute 'param1'

        my_dict.nested = DictDotAttrAccess(nested_dict)
        print(my_dict)  # {'param1': 1000, 'nested': {'param1': 'nested GOOD too'}}
        print(my_dict.nested.param1)  # 'nested GOOD too'

    ОСОБЕННОСТИ:
        +++И ТАК РАБОТАЛО
            1. +BLANK
                my_dict = DictDotAttrAccess()
                print(my_dict)  # {}

            2. +BOOL
                my_dict = DictDotAttrAccess()
                print(bool(my_dict))  # False

                my_dict = DictDotAttrAccess({'p1': '111'})
                print(my_dict)  # {'p1': '111'}
                print(bool(my_dict))  # True

            3. +IN LIST
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                print("p1" in my_dict)  # True

            5. +при обращении к НЕCУЩЕСТВУЮЩЕМУ - выдает None! без добавления параметра, а после присвоения - параметр создается
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                print(my_dict.p2)  #None
                print(my_dict)  #{'p1': '111'}
                my_dict.p2 = 222
                print(my_dict)  #{'p1': '111', 'p2': 222}

            6. +UPDATE
                my_dict= DictDotAttrAccess()
                print(my_dict)  # {}

                my_dict.update({1:1})
                print(my_dict)  # {1: 1}

            6. +POP
                my_dict = DictDotAttrAccess({1:1})

                my_dict.pop(1)
                print(my_dict)  # {}

            8. +ITEMS()
                my_dict = DictDotAttrAccess({'p1': '111'})
                print(my_dict.items())  #dict_items([('p1', '111')])

            10. +ИТЕРИРОВАТЬ
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                for i in my_dict:
                    print(i, my_dict.get(i))    # p1 111

            11. +КОПИЮ
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                my_dict2 = copy.copy(my_dict)
                print(my_dict2)  # {'p1': '111'}


        ---НЕ ПОЛУЧИЛОСЬ РЕАЛИЗОВАТЬ
            20. замену для __MISSING__
                def __missing__(self, key):
                    return "HELLO"

                __missing__ = lambda: "HELLO"
                __missing__ = FUNC_LINK_LAMBDA_REPEATER
    """
    # -----------------------------------------------------------------------------------------------------------------
    def __getitem__(self, item: Any) -> Any | NoReturn:
        result = self.get(item)         # thats wrong with NESTING! not will working with saving results!!!
        if isinstance(result, dict):
            result = DictDots(result)

        return result

    # -----------------------------------------------------------------------------------------------------------------
    def __getattr__(self, item: str) -> Any | None:
        return self[item]

    def __setattr__(self, item: str, value: Any) -> None | NoReturn:
        self[item] = value

    def __delattr__(self, item: str | Any) -> None:
        del self[item]


# =====================================================================================================================
if __name__ == '__main__':
    pass


# =====================================================================================================================



from typing import *

from PyQt5.QtCore import *
from PyQt5.QtGui import *

from base_aux.funcs import *
from base_aux.classes import *
from base_aux.pyqt import TableModelTemplate


# =====================================================================================================================
class TpTableModel(TableModelTemplate):
    DATA: "TpMultyDutBase"
    HEADERS: "Headers"

    # AUX -------------------------------------------
    open__settings: Optional[bool] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        class Headers(BreederStrStack):
            TESTCASE: int = 0
            SKIP: None = None
            ASYNC: None = None
            STARTUP_CLS: None = None
            DUTS: BreederStrSeries = BreederStrSeries(None, self.DATA.DEVICES__BREEDER_CLS.COUNT)
            TEARDOWN_CLS: None = None
            # FIXME: need resolve COUNT over DevicesIndexed!!!

        self.HEADERS = Headers()

    def rowCount(self, parent: QModelIndex = None, *args, **kwargs) -> int:
        return len(self.DATA.TCS__CLS)

    def columnCount(self, parent: QModelIndex = None, *args, **kwargs) -> int:
        return self.HEADERS.count()

    # def headerData(self, section: Any, orientation: Qt.Orientation, role: int = Qt.DisplayRole) -> str:
    #     if role == Qt.DisplayRole:
    #         # ------------------------------
    #         if orientation == Qt.Horizontal:
    #             return self.HEADERS[section]
    #
    #         # ------------------------------
    #         if orientation == Qt.Vertical:
    #             return str(section + 1)

    def flags(self, index: QModelIndex) -> int:
        # PREPARE -----------------------------------------------------------------------------------------------------
        col = index.column()
        row = index.row()

        # -------------------------------------------------------------------------------------------------------------
        flags = super().flags(index)

        if col in [self.HEADERS.SKIP, self.HEADERS.ASYNC] or col in self.HEADERS.DUTS:
            flags |= Qt.ItemIsUserCheckable
            # flags |= Qt.ItemIsSelectable
        else:
            # flags -= Qt.ItemIsSelectable
            pass

        # clear SELECTABLE ---------

        return flags

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        # PREPARE -----------------------------------------------------------------------------------------------------
        col = index.column()
        row = index.row()
        tc_cls = list(self.DATA.TCS__CLS)[row]

        dut = None
        tc_inst = None
        if col in self.HEADERS.DUTS:
            index = col - self.HEADERS.DUTS.START_OUTER
            dut = self.DATA.DEVICES__BREEDER_CLS.LIST__DUT[index]
            tc_inst = tc_cls.TCS__LIST[index]

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.DisplayRole:
            if col == self.HEADERS.TESTCASE:
                return f"{tc_cls.NAME}"
            if col == self.HEADERS.SKIP:
                return '+' if tc_cls.SKIP else '-'
            if col == self.HEADERS.ASYNC:
                return '+' if tc_cls.ASYNC else '-'

            # STARTUP -------------------
            if col == self.HEADERS.STARTUP_CLS:
                group_name = tc_cls.MIDDLE_GROUP__NAME or ""
                if tc_cls.result__startup_cls is None:
                    return group_name
                elif bool(tc_cls.result__startup_cls) is True:
                    return f'+{group_name}'
                elif bool(tc_cls.result__startup_cls) is False:
                    return f'-{group_name}'

            # DUTS -------------------
            if col in self.HEADERS.DUTS:
                if tc_inst:
                    if tc_inst.result is None:
                        return ""
                    else:
                        if bool(tc_inst.result):
                            return "PASS"
                        else:
                            return "FAIL"

            # TEARDOWN -------------------
            if col == self.HEADERS.TEARDOWN_CLS:
                if tc_cls.result__teardown_cls is None:
                    return
                elif bool(tc_cls.result__teardown_cls) is True:
                    return '+'
                elif bool(tc_cls.result__teardown_cls) is False:
                    return '-'

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.ToolTipRole:
            if col == self.HEADERS.TESTCASE:
                return f"{tc_cls.DESCRIPTION}"
            elif col in self.HEADERS.DUTS and tc_inst:
                return f"{tc_inst.result}"

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.TextAlignmentRole:
            """
            VARIANTS ALIGN
            --------------
            not exists NAME!!!} = 0         # (LEFT+TOP) [[[[[[[[DEFAULT IS [LEFT+TOP]]]]]]]]]
            
            AlignLeft=AlignLeading = 1      # LEFT(+TOP)
            AlignRight=AlignTrailing = 2    # RIGHT(+TOP)

            AlignTop = 32       # TOP(+LEFT)
            AlignBottom = 64    # BOT(+LEFT)

            AlignHCenter = 4    # HCENTER(+TOP)
            AlignVCenter = 128  # VCENTER(+LEFT)
            AlignCenter = 132   # VCENTER+HCENTER

            # =====MAYBE DID NOT FIGURED OUT!!!
            AlignAbsolute = 16      # (LEFT+TOP) == asDEFAULT
            AlignBaseline = 256     # (LEFT+TOP) == asDEFAULT

            AlignJustify = 8        # (LEFT+TOP) == asDEFAULT

            AlignHorizontal_Mask = 31   # TOP+RIGHT
            AlignVertical_Mask = 480    # LEFT+VCENTER
            """
            if col == self.HEADERS.TESTCASE:
                return Qt.AlignVCenter
            else:
                return Qt.AlignCenter

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.TextColorRole:
            if tc_cls.SKIP:
                return QColor('#a2a2a2')

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.BackgroundColorRole:
            if tc_cls.SKIP:
                return QColor('#e2e2e2')

            # ACTIVE ---------------------
            if col == self.HEADERS.TESTCASE:
                if self.DATA.tc_active == tc_cls:
                    return QColor("#FFFF50")

            # STARTUP -------------------
            if col == self.HEADERS.STARTUP_CLS:
                if tc_cls.result__startup_cls is None:
                    return
                elif bool(tc_cls.result__startup_cls) is True:
                    return QColor("#50FF50")
                elif bool(tc_cls.result__startup_cls) is False:
                    return QColor("#FF5050")

            # DUTS -------------------
            if col in self.HEADERS.DUTS:
                if tc_inst.skip_tc_dut or dut.SKIP or not dut.DEV_FOUND:
                    return QColor('#e2e2e2')
                elif tc_inst.result__startup is not None and not bool(tc_inst.result__startup):
                    return QColor("#FFa0a0")
                elif tc_inst.isRunning():
                    return QColor("#FFFF50")
                elif bool(tc_inst.result) is True:
                    return QColor("#00FF00")
                elif bool(tc_inst.result) is False:
                    if tc_inst.result is None:
                        return
                    else:
                        return QColor("#FF5050")
                # elif

            # TEARDOWN -------------------
            if col == self.HEADERS.TEARDOWN_CLS:
                if tc_cls.result__teardown_cls is None:
                    return
                elif bool(tc_cls.result__teardown_cls) is True:
                    return QColor("#50FF50")
                elif bool(tc_cls.result__teardown_cls) is False:
                    return QColor("#FF5050")

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.CheckStateRole:
            if self.open__settings:
                if col == self.HEADERS.SKIP:
                    if tc_cls.SKIP:
                        return Qt.Checked
                    else:
                        return Qt.Unchecked
                if col == self.HEADERS.ASYNC:
                    if tc_cls.ASYNC:
                        return Qt.Checked
                    else:
                        return Qt.Unchecked
                if col in self.HEADERS.DUTS:
                    if not tc_inst.SKIP and not dut.SKIP:
                        if tc_inst.skip_tc_dut:
                            return Qt.Unchecked
                        else:
                            return Qt.Checked

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.FontRole:
            if tc_cls == self.DATA.tc_active:
                # QFont("Arial", 9, QFont.Bold)
                font = QFont()

                font.setBold(True)
                # font.setItalic(True)

                # font.setOverline(True)  # надчеркнутый
                # font.setStrikeOut(True)  # зачеркнутый
                font.setUnderline(True)  # подчеркнутый

                # не понял!! --------------------
                # font.setStretch(5)
                # font.setCapitalization()

                return font

    def setData(self, index: QModelIndex, value: Any, role: int = None) -> bool:
        # PREPARE -----------------------------------------------------------------------------------------------------
        row = index.row()
        col = index.column()
        tc_cls = list(self.DATA.TCS__CLS)[row]

        dut = None
        tc_inst = None
        if col in self.HEADERS.DUTS:
            index = col - self.HEADERS.DUTS.START_OUTER
            dut = self.DATA.DEVICES__BREEDER_CLS.LIST__DUT[index]
            tc_inst = tc_cls(index=index)

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.CheckStateRole:
            if col == self.HEADERS.SKIP:
                tc_cls.SKIP = value == Qt.Checked

            if col == self.HEADERS.ASYNC:
                tc_cls.ASYNC = value == Qt.Checked

            if col in self.HEADERS.DUTS:
                if tc_inst:
                    tc_inst.skip_tc_dut = value == Qt.Unchecked

        # FINAL -------------------------------------------------------------------------------------------------------
        self._data_reread()
        return True


# =====================================================================================================================



from typing import *


# =====================================================================================================================
# TODO: use as ValueExplicit and VariantsExplicit(*args)
class Explicit:
    """
    GOAL
    ----
    1. solve NONE-VALUE as ambiguity by simple way
    2. show/pass explicitly the EXACT VALUE like None/[]/()

    RULES (when apply as funcResult)
    -----
    return object if you get final result!
    return None if there are any errors withing execution

    OLD ---- WHY: NamedTuple
    ----------------
    cause of we need to be able to compare different objects by values.
    maybe we need just add __eq__ method instead of it!!!

    USAGE
    -----
    to get exact value - just CALL instance! dont use VALUE access by attribute!

        from funcs import *

        def func(a, b) -> Optional[Explicit]:
            if a in b:
                return Explicit(a)
            else:
                return

        result = func("None", [None, ])
        assert result is None

        result = func(None, [None, ])
        assert result == Explicit(None)

        if result:
            print(result())         # None
    """
    __VALUE: Any   # dont use VALUE access by attribute

    def __init__(self, source: Any = None):
        self.__VALUE = source

    def __call__(self, *args, **kwargs) -> Any:
        return self.__VALUE

    def __str__(self):
        return f"{self.__class__.__name__}({self.__VALUE})"

    # -----------------------------------------------------------------------------------------------------------------
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__VALUE == other()
        else:
            return self.__VALUE == other


# =====================================================================================================================
class Default(Explicit):
    """
    GOAL
    ----
    just a derivative for Explicit!
    use in funcs as explicitly define for not passed params!

    USAGE
    -----
    1. AUTO_WORK (preferred)
        when get Default instance - call it to get source value!
        if get Default class - get None as value!

        def func(source = Default(1)):
            if isinstance(source, Default):
                source = source()

    2. MANUAL_WORK (not preferred)
        you can apply inside func

        def func(source = Default):
            if source == Default:
                source = 1
    """
    pass


# =====================================================================================================================
TYPE__EXPLICIT = Type[Explicit] | Explicit
TYPE__DEFAULT = Type[Default] | Default


# =====================================================================================================================



from . import *
from typing import *
import json

from . import TYPE__PV_DICT


# =====================================================================================================================
class PrivateJson(PrivateBase):
    FILENAME: str = "pv.json"

    def get_dict(self) -> Union[TYPE__PV_DICT, NoReturn]:
        """
        section only in first level!
        """
        try:
            json_data = json.loads(self._text)
        except Exception as exx:
            msg = f"[CRITICAL] incorrect format file!\n{exx!r}"
            print(msg)
            raise exx

        if self.SECTION:
            json_data = json_data.get(self.SECTION)

        if json_data:
            return json_data
        else:
            msg = f"[CRITICAL] NO [{self.SECTION=} in {self.filepath=}]\n"
            msg += self._text
            print(msg)


# =====================================================================================================================
class PrivateAuthJson(PrivateAuth, PrivateJson):
    pass


class PrivateTgBotAddressJson(PrivateTgBotAddress, PrivateJson):
    pass


# =====================================================================================================================



from typing import *
import subprocess
import time


# =====================================================================================================================
TYPE__CMD = Union[str, tuple[str, float | None]]
TYPE__CMDS = Union[TYPE__CMD, list[TYPE__CMD]]


# =====================================================================================================================
class Exx_CliNotAvailable(Exception):
    """CLI exx because of cli can not be available due to have not some of required commands.
    """


class Exx_CliTimeout(Exception):    #use direct subprocess.TimeoutExpired??? NO!!!
    """CLI exx because of timeout expired
    """


class Exx_CliRetcode(Exception):
    """CLI exx because of returnCode is not zero.
    """


class Exx_CliStderr(Exception):
    """CLI exx because of stderr have any data
    """


# =====================================================================================================================
class CliCmd:
    # TODO: ADD
    pass


# =====================================================================================================================
class CliUser:
    """Class which directly send commands to OS terminal

    :ivar TIMEOUT: default timeout for execution process
        if timeout expired and process still not finished - raise exx
    :ivar _last_sp: Popen object
    :ivar _last_exx_timeout: saved Exx_CliTimeout if it was
    :ivar last_cmd: command for execution
    :ivar last_stdout: full stdout data
    :ivar last_stderr: full stderr data
    :ivar last_retcode: returned code if process was last_finished
    :ivar last_finished: flag shows if command process was finished
        useful if execute in threading.

    :ivar CMDS_REQUIRED: commands for cli_check_available
        dict with NAME - exact commands which will send into terminal in order to check some utility is installed,
        VALUE - message if command get any error

        RECOMMENDATIONS
        --------------
        1. --HELP never works as expected! always timedout!!!!
            [#####################ERROR#####################]
            self.last_cmd='STM32_Programmer_CLI --help'
            self.last_duration=2.029675
            self.last_finished=True
            self.last_finished_success=False
            self.last_retcode=None
            --------------------------------------------------
            self.last_stdout=
            --------------------------------------------------
            self.last_stderr=
            --------------------------------------------------
            self._last_exx_timeout=Exx_CliTimeout("TimeoutExpired('STM32_Programmer_CLI --help', 2)")
            ==================================================
            [ERROR] cmd NOT available [STM32_Programmer_CLI --help]
            ==================================================

        2. DIRECT SIMPLE CLI COMMAND AS UTILITY_NAME.EXE without any parameter MAY NOT WORK!!! may timedout! implied the same as HELP parameter!
            [#####################ERROR#####################]
            self.last_cmd='STM32_Programmer_CLI'
            self.last_duration=2.022585
            self.last_finished=True
            self.last_finished_success=False
            self.last_retcode=None
            --------------------------------------------------
            self.last_stdout=
            --------------------------------------------------
            self.last_stderr=
            --------------------------------------------------
            self._last_exx_timeout=Exx_CliTimeout("TimeoutExpired('STM32_Programmer_CLI', 2)")
            ==================================================

        3. use --VERSION! instead! - seems work fine always!

    :exception Exx_CliNotAvailable:
    :exception Exx_CliTimeout:
    :exception Exx_CliRetcode:
    :exception Exx_CliStderr:
    """
    # SETTINGS ------------------------------------
    TIMEOUT: Optional[float] = 2
    RAISE: Optional[bool] = None

    CMDS_REQUIRED: Optional[Dict[str, Optional[str]]] = None

    _buffer_indent: str = "\t|"

    # VALUES --------------------------------------
    _last_sp: Optional[subprocess.Popen] = None

    counter: int = 0
    counter_in_list: int = 0

    last_duration: float = 0
    last_cmd: str = ""
    last_stdout: str = ""         # USE ONLY "" AS DEFAULT!!!
    last_stderr: str = ""         # USE ONLY "" AS DEFAULT!!!
    last_exx_timeout: Optional[Exx_CliTimeout] = None
    last_retcode: Optional[int] = None
    last_finished: Optional[bool] = None

    # init ------------------------------------------------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._reinit_values()       # need to keep always one values!

        if not self.cli_check_available():
            raise Exx_CliNotAvailable

    def _reinit_values(self) -> None:
        self._last_sp = None

        self.last_duration = 0
        self.last_cmd = ""
        self.last_stdout = ""
        self.last_stderr = ""
        self.last_exx_timeout = None
        self.last_retcode = None
        self.last_finished = None

    @property
    def last_finished_success(self) -> Optional[bool]:
        """check if process last_finished correct

        :return: True if last_finished with no errors
        """
        return all([self.last_finished, self.last_retcode == 0, not self.last_stderr, not self.last_exx_timeout])

    # SEND ------------------------------------------------------------------------------------------------------------
    def send(
            self,
            cmd: TYPE__CMDS,
            timeout: Optional[float] = None,
            till_first_true: Optional[bool] = None,
            _raise: Optional[bool] = None,
            _use_counter_list: Optional[bool] = None,
            print_all_states: Optional[bool] = None
    ) -> Union[bool, NoReturn]:
        """execute CLI command in terminal

        :param cmd: commands for execution
        :param timeout: use special timeout step_result_enum, instead of default, for cms_list will used as cumulated!
            # TODO: need decide about cumulation! more preferable apply as default! -NO its OK!
            so if you use list - apply timeout for CUMULATED! and dont pass it if no need cumulation!
            if single - apply as single!
        :param till_first_true: useful for detection or just multiPlatform usage
        :param _raise: if till_first_true=True it will not work (return always bool in this case)!!!
        :param _use_counter_list: DONT USE! internal flag for counter_in_list
        :param print_all_states: all or only Failed
        """
        # CMDS LIST ---------------------------------------------------------------------------------------------------
        if isinstance(cmd, list):
            self.counter_in_list = 0
            time_start = time.time()
            result = True

            _raise_list = _raise
            if till_first_true:
                _raise_list = False

            for cmd_item in cmd:
                self.counter_in_list += 1
                time_passed = time.time() - time_start
                try:
                    timeout = timeout - time_passed
                except:
                    pass

                result = self.send(cmd=cmd_item, timeout=timeout, _raise=_raise_list, _use_counter_list=True)

                if till_first_true:
                    if result:
                        return True
                else:
                    if not result:
                        msg = f"[ERROR] {cmd_item=} in full sequence {cmd=}"
                        print(msg)
                        return False

            return result

        # params reinit -----------------------------------------------------------------------------------------------
        self._reinit_values()
        if not _use_counter_list:
            self.counter_in_list = 0

        # params apply ------------------------------------------------------------------------------------------------
        msg = f"[CLI_SEND] [{cmd}]"
        print(msg)

        if isinstance(cmd, tuple) and len(cmd) == 2:
            cmd, timeout_i = cmd
            if timeout_i is not None:
                timeout = timeout_i

        self.last_cmd = cmd
        self.counter += 1

        if timeout is None:
            timeout = self.TIMEOUT
        if _raise is None:
            _raise = self.RAISE

        # work --------------------------------------------------------------------------------------------------------
        if timeout < 0:
            msg = f"{timeout=} is sub zero"
            self.last_exx_timeout = Exx_CliTimeout(msg)
            print(msg)
            if _raise:
                raise Exx_CliTimeout(msg)
            else:
                return False

        # todo: check for linux! encoding is not necessary!
        self._last_sp = subprocess.Popen(args=cmd, text=True, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="cp866")    #, encoding="cp866"

        # WORK --------------------------------
        time_start = time.time()
        lines = []
        while self._last_sp.poll() is None:
            self.last_duration = round(time.time() - time_start, 3)
            if timeout < self.last_duration:
                self._last_sp.kill()
                self.last_exx_timeout = Exx_CliTimeout()
                break
            line = self._last_sp.stdout.readline()
            if line != "":
                lines.append(line)
                print(".", end="")
            # print(f"[{repr(line)}]")

        if lines:
            self.last_stdout = "".join(lines)
            print()

        self.last_stderr = self._last_sp.stderr.read()
        self.last_retcode = self._last_sp.returncode
        self.last_finished = True
        if print_all_states or not self.last_finished_success:
            self.print_state()

        if _raise:
            if self.last_exx_timeout:
                raise self.last_exx_timeout
            if self.last_retcode != 0:
                raise Exx_CliRetcode(self.last_retcode)
            if self.last_stderr:
                raise Exx_CliStderr(self.last_stderr)

        return self.last_finished_success

    # AUXILIARY -------------------------------------------------------------------------------------------------------
    def print_state(self) -> None:
        print(f"="*50)
        if not self.last_finished_success:
            print(f"[{'#'*21}ERROR{'#'*21}]")

        print(f"{self.counter=}")
        print(f"{self.counter_in_list=}")

        print(f"{self.last_cmd=}")
        print(f"{self.last_duration=}")
        print(f"{self.last_finished=}")
        print(f"{self.last_finished_success=}")
        print(f"{self.last_retcode=}")
        print(f"-" * 50)
        print("self.last_stdout=")
        if self.last_stdout:
            for line in self.last_stdout.split("\n"):
                print(f"{self._buffer_indent}{line!r}")
        print(f"-"*50)
        print("self.last_stderr=")
        if self.last_stderr:
            for line in self.last_stderr.split("\n"):
                print(f"{self._buffer_indent}{line!r}")
        print(f"-" * 50)
        print(f"{self.last_exx_timeout=}")
        print(f"="*50)

    def cli_check_available(self) -> bool:
        """check list of commands which will show that further work will executable and your environment is ready.

        Useful because commands uwually depends on installed programs and OS.
        so if you want to be sure of it on start point - run it!
        """
        if self.CMDS_REQUIRED:
            for cmd, error_msg in self.CMDS_REQUIRED.items():
                if not self.send(cmd, _raise=False):
                    msg = f"[ERROR] cmd NOT available [{cmd}]"
                    print(msg)
                    self.print_state()
                    return False
        return True


# =====================================================================================================================



from typing import *
from base_aux.objects import *
from . import Exx__GetattrPrefix_RaiseIf

from .valid_0_aux import ValidAux
from .attr_1_aux import AttrAux


# =====================================================================================================================
class GetattrPrefixInst(AttrAux):
    """
    this is just a Base!

    EXAMPLES
    --------
    see GetattrPrefixInst_RaiseIf with tests
    """
    _GETATTR__PREFIXES: list[str] = []

    def __getattr__(self, item: str) -> Any | Callable | NoReturn:
        """
        SHARING PARAMS BETWEEN CALLABLE PREFIX/ITEM
        -------------------------------------------
        you can not use params! and not use callables (static attributes are available)!
        args - all - goes for ITEM
        kwargs - all uppercase - goes for PREFIX (after changing by lower())
        kwargs - others - goes for ITEM directly (without changing case)!

        if you provide not existed args/kwargs - you will get direct corresponding exception like "print(hello=1) -> TypeError: 'hello' is an invalid keyword argument for print() "

        NOTE
        ----
        0. CaseInSensitive!
        1. prefix - callable with first parameter as catching item_value (may be callable or not).
        2.You always need to CALL prefixed result! even if you access to not callable attribute!
        """
        for prefix in self._GETATTR__PREFIXES:
            prefix_original = self._attr_anycase__find(prefix, self)
            if not prefix_original:
                continue
            prefix_meth = getattr(self, prefix_original)

            if item.lower() == prefix.lower():
                return lambda *prefix_args, **prefix_kwargs: ValidAux.get_result_or_raise(
                    source=prefix_meth,
                    args=prefix_args,
                    kwargs=prefix_kwargs
                )

            if prefix_original and item.lower().startswith(prefix.lower()):
                item_short = item[len(prefix):]
                item_value = self._getattr_anycase(item_short, self)

                return lambda *meth_args, **meth_kwargs: ValidAux.get_result_or_raise(
                    source=prefix_meth,
                    args=[ValidAux.get_result_or_raise(source=item_value, args=meth_args, kwargs={k:v for k,v in meth_kwargs.items() if not k.isupper()}), ],
                    kwargs={k.lower():v for k,v in meth_kwargs.items() if k.isupper()}
                )

        raise AttributeError(item)


# =====================================================================================================================
class GetattrPrefixInst_RaiseIf(GetattrPrefixInst):
    """
    RULES
    -----
    """
    _GETATTR__PREFIXES = ["raise_if__", "raise_if_not__"]

    # -----------------------------------------------------------------------------------------------------------------
    def raise_if__(self, source: Any, _reverse: bool | None = None, comment: str = "") -> None | NoReturn:
        _reverse = _reverse or False
        result = ValidAux.get_result_or_exx(source=source)
        if TypeChecker.check__exception(result) or bool(result) != bool(_reverse):
            raise Exx__GetattrPrefix_RaiseIf(f"[raise_if__/{_reverse=}]met conditions ({source=}/{comment=})")

    def raise_if_not__(self, source: Any, comment: str = "") -> None | NoReturn:
        return self.raise_if__(source=source, _reverse=True, comment=comment)


# =====================================================================================================================
def _example():
    class GetattrPrefixInst_RaiseIf_data(GetattrPrefixInst_RaiseIf):
        ATTR0 = 0
        ATTR1 = 1
        def METH(self, arg1=1, arg2=2):
            return arg1 == arg2

    # ON ATTRIBUTE --------------------
    # apply prefix+item
    GetattrPrefixInst_RaiseIf_data().raise_if__attr0()  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__attr1()  # Exx__GetattrPrefix_RaiseIf

    # ON METH -------------------------
    # args - one
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(0)  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1)  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(2)  # Exx__GetattrPrefix_RaiseIf

    # args - several
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1,2)  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(2,2)  # Exx__GetattrPrefix_RaiseIf

    # kwargs
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1, arg2=2)  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(2, arg2=2)  # Exx__GetattrPrefix_RaiseIf

    # send kwarg for PREFIX ----------
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1, arg2=2, comment="WRONG!")    # RAISE comment argument invalid for meth()!!!
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1, arg2=2, COMMENT="CORRECT!")  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(2, arg2=2, COMMENT="CORRECT!")  # Exx__GetattrPrefix_RaiseIf


# =====================================================================================================================
